# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.sketchpad_canvas import SketchpadCanvas
from libqtopensesame.widgets.sketchpad_widget import SketchpadWidget
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame import sketchpad_elements
from libqtopensesame.validators import DurationValidator


class Feedpad:

    r"""Controls for the sketchpad item."""
    help_url = u'manual/stimuli/visual'
    lazy_init = True

    def init_edit_widget(self):
        r"""Initializes the widget."""
        QtPlugin.init_edit_widget(self, False)
        self.canvas = SketchpadCanvas(self)
        self.sketchpad_widget = SketchpadWidget(self)
        self.add_widget(self.sketchpad_widget)
        self.auto_add_widget(self.sketchpad_widget.ui.edit_duration,
                             u'duration')
        self.sketchpad_widget.ui.edit_duration.setValidator(
            DurationValidator(self))
        self.first_refresh = True
        self._lock = False

    def apply_script_changes(self):
        """See qtitem."""
        # Without this hack, the script changes will automatically trigger a
        # draw(), which triggers a call to apply_edit_changes, which updates the
        # script. To avoid this feedback loop, temporarily disable apply edit
        # changes.
        f = self.apply_edit_changes
        self.apply_edit_changes = lambda: None
        QtPlugin.apply_script_changes(self)
        self.apply_edit_changes = f

    def element_module(self):
        """
        returns:
                desc:	The module that contains the sketchpad elements.
                type:	module
        """
        return sketchpad_elements

    def set_focus(self):
        r"""Allows the item to focus the most important widget."""
        self.sketchpad_widget.ui.edit_duration.setFocus()

    def edit_widget(self):
        r"""Updates the widget."""
        QtPlugin.edit_widget(self)
        self.sketchpad_widget.initialize()
        self.sketchpad_widget.draw()
        if self.first_refresh:
            self.sketchpad_widget.center()
            self.sketchpad_widget.zoom_fit()
            self.first_refresh = False

    def add_element(self, element):
        r"""Adds an element to the sketchpad.

        Parameters
        ----------
        element : base_element
            The element to add.
        """
        if element is None:
            return
        self.elements.append(element)
        self.draw()
        for element in self.elements:
            element.select(False)
        element.select()

    def selected_elements(self):
        r"""Gets selected elements.

        Returns
        -------
        list
            A list of selected elements.
        """
        l = []
        for element in self.elements:
            if element.selected:
                l.append(element)
        return l

    def remove_elements(self, elements):
        r"""Removes elements.

        Parameters
        ----------
        elements : list
            A list of elements to be removed.
        """
        for element in elements:
            self.elements.remove(element)

    def move_elements(self, elements, dx=0, dy=0):
        r"""Moves elements.

        Parameters
        ----------
        elements : list
            A list of elements to be moved.
        dx : int, float, optional
            A horizontal displacement.
        dy : int, float, optional
            A vertical displacement
        """
        for element in elements:
            element.move(dx, dy)

    def topleft_elements_pos(self, elements):
        r"""Gets the top-left position of a set of elements.

        Parameters
        ----------
        elements : list
            A list of elements.

        Returns
        -------
        tuple.
            An (x,y) tuple with the top-left position.
        """
        lx = []
        ly = []
        for element in elements:
            x, y = element.get_pos()
            lx.append(x)
            ly.append(y)
        return min(lx), min(ly)

    def set_elements_pos(self, elements, x=0, y=0):
        r"""Sets position of elements.

        Parameters
        ----------
        elements : list
            A list of elements to be positioned.
        x : int, float, optional
            A horizontal position.
        y : int, float, optional
            A vertical position
        """
        mx, my = self.topleft_elements_pos(elements)
        for element in elements:
            dx, dy = element.get_pos()
            dx -= mx
            dy -= my
            element.set_pos(x+dx, y+dy)

    def min_z_index(self):
        r"""Gets the z-index of the top item (i.e. in front of all other
        items).

        Returns
        -------
        int
            A z-index.
        """
        return min([
            element.z_index
            for element in self.elements
            if isinstance(element.z_index, (int, float))
        ])

    def max_z_index(self):
        r"""Gets the z-index of the bottom item (i.e. below all other items).

        Returns
        -------
        int
            A z-index.
        """
        return max([
            element.z_index
            for element in self.elements
            if isinstance(element.z_index, (int, float))
        ])

    @property
    def current_color(self):
        return self.sketchpad_widget.current_color

    @property
    def current_penwidth(self):
        return self.sketchpad_widget.current_penwidth

    @property
    def current_arrow_head_width(self):
        return self.sketchpad_widget.current_arrow_head_width

    @property
    def current_arrow_body_width(self):
        return self.sketchpad_widget.current_arrow_body_width

    @property
    def current_arrow_body_length(self):
        return self.sketchpad_widget.current_arrow_body_length

    @property
    def current_scale(self):
        return self.sketchpad_widget.current_scale

    @property
    def current_rotation(self):
        return self.sketchpad_widget.current_rotation

    @property
    def current_show_if(self):
        return self.sketchpad_widget.current_show_if

    @property
    def current_fill(self):
        return self.sketchpad_widget.current_fill

    @property
    def current_center(self):
        return self.sketchpad_widget.current_center

    @property
    def current_font_family(self):
        return self.sketchpad_widget.current_font_family

    @property
    def current_font_size(self):
        return self.sketchpad_widget.current_font_size

    @property
    def current_font_bold(self):
        return self.sketchpad_widget.current_font_bold

    @property
    def current_font_italic(self):
        return self.sketchpad_widget.current_font_italic

    @property
    def current_html(self):
        return self.sketchpad_widget.current_html

    @property
    def draw(self):
        return self.sketchpad_widget.draw

    @property
    def set_cursor_pos(self):
        return self.sketchpad_widget.set_cursor_pos

    @property
    def graphics_view(self):
        return self.sketchpad_widget.ui.graphics_view

    @property
    def zoom_diff(self):
        return self.sketchpad_widget.zoom_diff

    @property
    def show_element_settings(self):
        return self.sketchpad_widget.show_element_settings

    @property
    def select_pointer_tool(self):
        return self.sketchpad_widget.select_pointer_tool


# Alias for backwards compatibility
feedpad = Feedpad



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import libopensesame.python_workspace_api as api
from libopensesame.exceptions import AbortCoroutines
import types
from openexp.canvas_elements import ElementFactory
from libopensesame.widgets.widget_factory import WidgetFactory
from libopensesame.base_python_workspace import BasePythonWorkspace


class PythonWorkspace(BasePythonWorkspace):

    """
    desc:
        Provides a workspace in which Python scripts can be executed. Generally,
        each experiment has one python_workspace object, which is used for all
        inline_script items, and other items that make allow users to execute
        Python code.

        The python_workspace object is dictionary-like, and allows you to get
        and set global variables in a dictionary way, e.g.:

            self.experiment.python_workspace[u'my_var'] = 10
    """
    def __init__(self, experiment):
        r"""Constructor.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        """
        super().__init__(experiment)
        api.experiment = experiment

    def init_globals(self):
        r"""Initialize the global workspace."""
        # By setting the __name__ global, the workspace will operate as a
        # module, so that e.g. import statements don't suffer from locality.
        self._globals.update({
            u'__name__': u'python_workspace',
            u'exp': self.experiment,
            u'var': self.experiment.var,
            u'pool': self.experiment.pool,
            u'items': self.experiment.items,
            u'clock': self.experiment._clock,
            u'log': self.experiment._log,
            u'responses': self.experiment._responses,
            u'data_files': self.experiment.data_files,
            u'AbortCoroutines': AbortCoroutines,
        })
        # All functions from the api modules are also loaded into the globals.
        # This way they can be called directly by name.
        api.set_aliases()
        for name, obj in api.__dict__.items():
            if self._is_api_object(obj):
                self._globals[name] = obj

    def _is_api_object(self, obj):

        if isinstance(obj, types.FunctionType):
            return True
        if isinstance(obj, type) and issubclass(
                obj, (WidgetFactory, ElementFactory)):
            return True
        return False


# Alias for backwards compatibility
python_workspace = PythonWorkspace



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad_elements._arrow import Arrow
from libopensesame.sketchpad_elements._circle import Circle
from libopensesame.sketchpad_elements._ellipse import Ellipse
from libopensesame.sketchpad_elements._fixdot import Fixdot
from libopensesame.sketchpad_elements._gabor import Gabor
from libopensesame.sketchpad_elements._image import Image
from libopensesame.sketchpad_elements._line import Line
from libopensesame.sketchpad_elements._noise import Noise
from libopensesame.sketchpad_elements._textline import Textline
from libopensesame.sketchpad_elements._rect import Rect

# Alias for backwards compatibility
arrow = Arrow
circle = Circle
ellipse = Ellipse
fixdot = Fixdot
gabor = Gabor
image = Image
line = Line
noise = Noise
textline = Textline
rect = Rect



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._polygon.polygon import Polygon
from openexp._canvas._element.xpyriment import XpyrimentElement
from expyriment.stimuli import Shape
from expyriment.misc.geometry import points2vertices, XYPoint


class Xpyriment(XpyrimentElement, Polygon):

    def prepare(self):

        left = min(x for x, y in self.vertices)
        right = max(x for x, y in self.vertices)
        top = min(y for x, y in self.vertices)
        bottom = max(y for x, y in self.vertices)
        pos = left + (right-left)//2, top + (bottom-top)//2
        self._stim = Shape(
            line_width=0 if self.fill else self.penwidth,
            position=self.to_xy(pos),
            colour=self.color.backend_color,
            anti_aliasing=self.ANTI_ALIAS
        )
        self._stim.add_vertices(points2vertices(
            [XYPoint(xy=self.to_xy(x, y)) for x, y in self.vertices]
        ))
        self._stim.preload()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad_elements._base_element import BaseElement


class Arrow(BaseElement):

    r"""An arrow element for the sketchpad."""
    def __init__(self, sketchpad, string):
        r"""Constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        string
            A definition string.
        """
        defaults = [
            (u'x1', None),
            (u'y1', None),
            (u'x2', None),
            (u'y2', None),
            (u'arrow_head_width', 30),
            (u'arrow_body_width', 0.5),
            (u'arrow_body_length', 0.8),
            (u'color', sketchpad.var.foreground),
            (u'penwidth', 1),
            (u'fill', True),
        ]
        super().__init__(sketchpad, string, defaults=defaults)

    def draw(self):
        r"""Draws the element to the canvas of the sketchpad."""
        properties = self.eval_properties()
        return self.canvas.arrow(properties[u'x1'], properties[u'y1'],
                                 properties[u'x2'], properties[u'y2'],
                                 color=properties[u'color'],
                                 penwidth=properties[u'penwidth'],
                                 head_width=properties[u'arrow_head_width'],
                                 body_width=properties[u'arrow_body_width'],
                                 body_length=properties[u'arrow_body_length'],
                                 fill=properties[u'fill'])


# Alias for backwards compatibility
arrow = Arrow



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.misc import snake_case
from libopensesame.exceptions import InvalidSketchpadElementScript, \
    ConditionalExpressionError


class BaseElement:
    r"""A base class from which all sketchpad elements are derived."""
    
    def __init__(self, sketchpad, string, defaults=[]):
        r"""Constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        string
            A definition string.
        defaults, optional
            A list with (name, default_value) tuples for all keywords.
        """
        self._type = snake_case(self.__class__.__name__)
        self.defaults = defaults + [
            (u'z_index', 0),
            (u'show_if', u'always'),
            (u'name', u'')
        ]
        self.sketchpad = sketchpad
        self.var = self.sketchpad.var
        self.from_string(string)

    @property
    def element_name(self):

        p = self.eval_properties()
        if not p[u'name']:
            return None
        return p[u'name']

    @property
    def canvas(self): return self.sketchpad.canvas

    @property
    def pool(self): return self.sketchpad.experiment.pool

    @property
    def name(self): return self.sketchpad.name

    @property
    def syntax(self): return self.sketchpad.syntax

    @property
    def experiment(self): return self.sketchpad.experiment

    @property
    def z_index(self):
        r"""Determines the drawing order of the elements. Elements with a
        higher z-index are drawn first, so they are at the bottom of the stack.

        Returns
        -------
        A z-index.
        """
        return self.properties[u'z_index']

    def draw(self):
        r"""Draws the element to the canvas of the sketchpad."""
        pass

    def from_string(self, s):
        r"""Parse a definition string for the element.

        Parameters
        ----------
        s
            A definition string.
        """
        cmd, arglist, kwdict = self.syntax.parse_cmd(s)
        if cmd != u'draw' or len(arglist) == 0 or arglist[0] != self._type:
            raise InvalidSketchpadElementScript(
                f'Invalid script: "{s}"')
        # First load the default values
        self.properties = {}
        for var, val in self.defaults:
            self.properties[var] = val
        # Parse the argument list. This is the old way, in which arguments where
        # passed by order.
        for i, val in enumerate(arglist[1:]):
            var = self.defaults[i][0]
            self.properties[var] = val
        # Now parse keywords
        self.properties.update(kwdict)
        # Check if all values that need to be specified have indeed been
        # specified.
        for var, val in self.properties.items():
            if val is None:
                raise InvalidSketchpadElementScript(
                    f'Required keyword {var} has not been specified in '
                    f'sketchpad element {self._type}')
        # Check if no non-existing keywords have been specified
        for var in self.properties.keys():
            valid = False
            for _var, _val in self.defaults:
                if _var == var:
                    valid = True
                    break
            if not valid:
                raise InvalidSketchpadElementScript(
                    f'The keyword {var} is not applicable to '
                    f'sketchpad element {self._type}')

    def valid_keyword(self, keyword):
        r"""Checks whether a particular keyword is valid for this element.

        Parameters
        ----------
        keyword
            A keyword.
        type
            unicode

        Returns
        -------
        bool
            True if keyword is valid.
        """
        for var, val in self.defaults:
            if var == keyword:
                return True
        return False

    def escape(self, val, quote=True):
        r"""Escapes and optionally quotes a value so that it can be safely
        inserted into a definition string. Everything except unicode is
        returned as is.

        Parameters
        ----------
        val : unicode, float, int
            The value to escape.
        quote : bool
            Indicates whether unicode strings should be quoted with double
            quotes.

        Returns
        -------
        unicode, int, float
            A value that can be safely inserted into a definition string.
        """
        if not isinstance(val, str):
            return val
        val = val.replace(u'\\', u'\\\\')
        val = val.replace(u'"', u'\\"')
        if quote:
            val = u'"%s"' % val
        return val

    def to_string(self):
        r"""Generates a string representation of the element.

        Returns
        -------
        unicode
            A string representation.
        """
        return self.syntax.create_cmd(u'draw', [self._type],
                                      {var: val for var, val in self.properties.items()
                                       if var != u'name' or val}
                                      )

    def eval_properties(self):
        r"""Evaluates all properties.

        Returns
        -------
        A new property dictionary.
        """
        properties = {}
        xc = self.var.width/2
        yc = self.var.height/2
        for var, val in self.properties.items():
            if var == u'text':
                round_float = True
            else:
                round_float = False
            val = self.sketchpad.syntax.auto_type(
                self.sketchpad.syntax.eval_text(val, round_float=round_float))
            properties[var] = val
        return properties

    def is_shown(self):
        r"""Determines whether the element should be shown, based on the show-
        if statement.

        Returns
        -------
        bool
            A bool indicating whether the element should be shown.
        """
        self.experiment.python_workspace[u'self'] = self.sketchpad
        bytecode = self.experiment.syntax.compile_cond(
            self.properties[u'show_if'])
        try:
            return self.experiment.python_workspace._eval(bytecode)
        except Exception as e:
            raise ConditionalExpressionError(
                'Error while evaluating show-if expression')


# Alias for backwards compatibility
base_element = BaseElement



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from openexp import backend
import opensesame_plugins
from libopensesame.plugin_manager import PluginManager
from libopensesame.var_store import VarStore
from libopensesame.item_store import ItemStore
from libopensesame.response_store import ResponseStore
from libopensesame.file_pool_store import FilePoolStore
from libopensesame.syntax import Syntax
from libopensesame.exceptions import UserAborted, ItemDoesNotExist, \
    InvalidOpenSesameScript
from libopensesame.item import Item
from libopensesame import misc, metadata
from libopensesame.item_stack import item_stack_singleton
from libopensesame.oslogging import oslogger
from libopensesame.py3compat import *
import os
import sys
import pickle
import time
import warnings
import gc


class Experiment(Item):

    """A special item that controls the flow of the experiment."""
    description = 'The main experiment item'

    def __init__(self, name='experiment', string=None, pool_folder=None,
                 experiment_path=None, fullscreen=False,
                 logfile='defaultlog.csv', subject_nr=0, workspace=None,
                 resources={}, heartbeat_interval=1):
        """Constructor. The experiment is created automatically be OpenSesame
        and you will generally not need to create it yourself.

        Parameters
        ----------
        name : str, unicode, optional
            The name of the experiment.
        string : str, unicode, NoneType, optional
            A string containing the experiment definition, the name of an
            OpenSesame experiment file, or `None` to create a blank experiment.
        pool_folder : str, unicode, NoneType, optional
            A specific folder to be used for the file pool, or `None` to use a
            new temporary folder.
        experiment_path : str, optional
            The path of the experiment file. This is the folder that the
            experiment is in, not the path to the experiment file.
        fullscreen : bool, optional
            Indicates whether the experiment should be executed in fullscreen.
        logfile : unicode, str, optional
            The logfile path.
        subject_nr : int, optional
            The subject number.
        workspace : python_workspace, NoneType, optional
            A `python_workspace` object to be used for executing custom Python
            code, or `None` to create a new workspace.
        resources : dict, optional
            A dictionary with names as keys and paths as values. This serves as
            a look-up table for resources.
        heartbeat_interval : int, float, optional
            A heartbeat interval in seconds, or <= 0 to disable heartbeats.
        """
        # Make sure the logger is started
        if not oslogger.started:
            oslogger.start()
            oslogger.debug('starting logger in experiment init')
        self.var = VarStore(self)
        self.pool = FilePoolStore(self, folder=pool_folder)
        self._responses = ResponseStore(self)
        # The _syntax and items objects may already have been created by
        # libqtopensesame.experiment.
        if not hasattr(self, '_syntax'):
            self._syntax = Syntax(self)
        if not hasattr(self, 'items'):
            self.items = ItemStore(self)
        if not hasattr(self, '_plugin_manager'):
            self._plugin_manager = PluginManager(opensesame_plugins)
        if workspace is None:
            from libopensesame.python_workspace import PythonWorkspace
            self._python_workspace = PythonWorkspace(self)
        else:
            self._python_workspace = workspace
        self.running = False
        self.heartbeat_interval = heartbeat_interval
        self.plugin_folder = 'plugins'
        self._start_response_interval = None
        self.cleanup_functions = []
        self.restart = False
        self.resources = resources
        self.paused = False
        self.output_channel = None
        self.reset()

        # Logfile parameters
        self._log = None
        self.logfile = logfile
        # A list of data files, which may include more than just the OpenSesame
        # logfile, for example if data is also recorded using some other method.
        self.data_files = []

        # This is some duplication of the option parser in qtopensesame,
        # but nevertheless keep it so we don't need qtopensesame
        self.debug = '--debug' in sys.argv or '-d' in sys.argv
        if string is not None:
            string = self.open(string)
        Item.__init__(self, name, self, string)
        # Default subject info
        self.set_subject(subject_nr)
        # Fullscreen needs to be set after the experiment has been parsed from
        # script, otherwise it will be overridden by the script.
        self.var.fullscreen = 'yes' if fullscreen else 'no'
        # Restore experiment path, which is either the full path (including
        # filename), only the folder, or None.
        if experiment_path is not None:
            if os.path.isfile(experiment_path):
                self.var.experiment_path, self.var.experiment_file = \
                    self.experiment_path, self.experiment_file = \
                    os.path.split(experiment_path)
            else:
                self.var.experiment_path = self.experiment_path \
                    = experiment_path
        # When a template is opened, it should not result in the experiment
        # path being defined, because it is intended to be saved as a new
        # experiment. Templates are identified by the is_template variable,
        # which is removed when the file is opened.
        if self.var.get('is_template', 0):
            self.experiment_path = self.var.experiment_path = None
            del self.var.is_template
        if not py3 and backend.backend_guess(self, 'sampler') == 'psycho':
            self.var.sampler_backend = 'legacy'
            oslogger.warning(
                'The psycho sampler backend is not compatible with Python 2. '
                'Falling back to legacy sampler backend'
            )

    @property
    def default_title(self):
        return 'New experiment'

    def reset(self):
        # Set default variables
        self.var.start = 'experiment'
        self.var.title = self.default_title
        self.var.round_decimals = 2
        self.var.form_clicks = 'no'
        self.var.disable_garbage_collection = 'yes'
        # Sound parameters
        self.var.sound_freq = 48000
        self.var.sound_sample_size = -16  # Negative values mean signed
        self.var.sound_channels = 2
        self.var.sound_buf_size = 1024
        # Default backend
        self.var.canvas_backend = 'psycho'
        # Display parameters
        self.var.width = 1024
        self.var.height = 768
        self.var.background = 'black'
        self.var.foreground = 'white'
        # Font parameters
        self.var.font_size = 18
        self.var.font_family = 'mono'
        self.var.font_italic = 'no'
        self.var.font_bold = 'no'
        self.var.font_underline = 'no'

    def module_container(self):
        """Specify the module that contains the item modules"""
        return 'libopensesame'

    def item_prefix(self):
        """A prefix for the plug-in classes, so that [prefix][plugin] class is
        used instead of the [plugin] class.
        """
        return ''

    def set_subject(self, nr):
        """Sets the subject number and parity (even/ odd). This function is
        called automatically when an experiment is started, so you do not
        generally need to call it yourself.

        Parameters
        ----------
        nr : int
            The subject nr.

        Examples
        --------
        >>> exp.set_subject(1)
        >>> print('Subject nr = %d' % exp.get('subject_nr'))
        >>> print('Subject parity = %s' % exp.get('subject_parity'))
        """
        # Set the subject nr and parity
        self.var.subject_nr = nr
        if nr % 2 == 0:
            self.var.subject_parity = 'even'
        else:
            self.var.subject_parity = 'odd'

    def read_definition(self, s):
        """Extracts a the definition of a single item from the string.

        Parameters
        ----------
        s: str
            The definition string.

        Returns
        -------
        tuple
            A (str, str) tuple with the full string minus the definition string
            and the definition string.
        """
        # Read the string until the end of the definition
        def_str = ''
        line = next(s, None)
        if line is None:
            return None, ''
        get_next = False
        while True:
            if len(line) > 0:
                if line[0] != '\t':
                    break
                else:
                    def_str += line + '\n'
            line = next(s, None)
            if line is None:
                break
        return line, def_str

    def from_string(self, string):
        """Reads the entire experiment from a string.

        Parameters
        ----------
        string
            The definition string.
        """
        self.var.clear(preserve=['experiment_path', 'experiment_file'])
        self.reset()
        self.comments = []
        oslogger.debug(u"building experiment")
        if string is None:
            return
        self.front_matter, string = self._syntax.parse_front_matter(string)
        if self.experiment.front_matter['API'].version[0] < 2:
            # Old experiment scripts were saved in ASCII, and require decoding
            # of U+XXXX unicode characters.
            string = self.syntax.from_ascii(string)
        s = iter(string.split('\n'))
        line = next(s, None)
        while line is not None:
            get_next = True
            l = self.syntax.split(line)
            if l:
                self.parse_variable(line)
                # Parse definitions
                if l[0] == u"define":
                    if len(l) != 3:
                        raise InvalidOpenSesameScript(
                            'Failed to parse definition', line=line)
                    item_type = l[1]
                    item_name = self.syntax.sanitize(l[2])
                    line, def_str = self.read_definition(s)
                    get_next = False
                    self.items.new(
                        _type=item_type,
                        name=item_name,
                        script=def_str,
                        allow_rename=False
                    )
            # Advance to next line
            if get_next:
                line = next(s, None)

    def transmit_workspace(self, **extra):
        """Sends the current workspace through the output channel. If there is
        no output channel, this function does nothing.

        Parameters
        ----------
        **extra : dict
            Any extra items in the workspace dict to be sent.
        """
        if self.output_channel is None:
            return
        d = self.python_workspace._globals.copy()
        d.update(extra)
        for key, value in d.copy().items():
            try:
                pickle.dumps(value)
            except:
                del d[key]
        self.output_channel.put(d)

    def set_output_channel(self, output_channel):
        """Sets the output channel, which is used to communicate the workspace
        between the experiment and the launch process (typically the GUI).

        Parameters
        ----------
        output_channel    The output object, which must support a `put` method.
        """
        if not hasattr(output_channel, 'put'):
            raise RuntimeError(f'Invalid output_channel: {output_channel}')
        self.output_channel = output_channel

    def run(self):
        """Runs the experiment."""
        # Save the date and time, and the version of OpenSesame
        self.var.datetime = safe_decode(
            time.strftime('%c'),
            enc=self.encoding,
            errors='ignore'
        )
        self.var.opensesame_version = metadata.__version__
        self.var.opensesame_codename = metadata.codename
        self.running = True
        self.init_random()
        self.init_display()
        self.init_clock()
        self.init_sound()
        self.init_log()
        self.python_workspace.init_globals()
        self.reset_feedback()
        self.init_heartbeat()
        oslogger.info(u"experiment started")
        if self.var.start in self.items:
            item_stack_singleton.clear()
            if self.var.disable_garbage_collection == 'yes':
                oslogger.info('disabling garbage collection')
                gc.disable()
            self.items.execute(self.var.start)
        else:
            raise ItemDoesNotExist(self.var.start)
        oslogger.info(u"experiment finished")
        self.end()

    def pause(self):
        """Pauses the experiment, sends the Python workspace to the GUI, and
        waits for the GUI to send a resume signal. This requires an output
        channel.
        """
        if self.paused:
            return

        from openexp.canvas import Canvas
        from openexp.canvas_elements import Text
        from openexp.keyboard import Keyboard

        self.paused = True
        self.transmit_workspace(__pause__=True)
        pause_canvas = Canvas(self)
        pause_canvas['msg'] = Text(
            'The experiment has been paused<br /><br />'
            'Press spacebar to resume<br />'
            'Press Q to quit'
        ).construct(pause_canvas)
        pause_keyboard = Keyboard(
            self,
            keylist=['space', 'q', 'Q'],
            timeout=0
        )
        pause_keyboard.show_virtual_keyboard()
        pause_canvas.show()
        try:
            while True:
                key, _time = pause_keyboard.get_key()
                if key == 'q' or key == 'Q':
                    pause_keyboard.show_virtual_keyboard(False)
                    raise UserAborted('The experiment was aborted')
                if key == 'space':
                    break
                time.sleep(.25)
        finally:
            self.paused = False
            self.transmit_workspace(__pause__=False)
        pause_keyboard.show_virtual_keyboard(False)
        pause_canvas['msg'].text = 'Resuming ...'
        pause_canvas.show()

    def cleanup(self):
        """Calls all the cleanup functions."""
        while len(self.cleanup_functions) > 0:
            func = self.cleanup_functions.pop()
            oslogger.debug(u"calling cleanup function")
            func()

    def end(self):
        """Nicely ends the experiment."""
        from openexp import sampler, canvas
        self.running = False
        try:
            self._log.close()
        except AttributeError:
            oslogger.error('missing or invalid log object')
        sampler.close_sound(self)
        canvas.close_display(self)
        self.cleanup()
        if not gc.isenabled():
            oslogger.info('enabling garbage collection')
            gc.enable()
        self.transmit_workspace(__finished__=True)

    def to_string(self):
        """
        Encodes the experiment into a string.

        Returns:
        A Unicode definition string for the experiment.
        """
        s = self._syntax.generate_front_matter()
        for var in self.var:
            s += self.variable_to_string(var)
        s += '\n'
        for _item in sorted(self.items):
            s += self.items[_item].to_string() + '\n'
        return s

    def save(self, path, overwrite=False, update_path=True):
        """Saves the experiment to file.

        Parameters
        ----------
        path : str, unicode
            The target file to save to.
        overwrite : bool, optional
            Indicates if existing files should be overwritten.
        update_path : bool, optional
            Indicates if the experiment_path attribute should be updated.

        Returns
        -------
        unicode, bool
            The path on successful saving or False otherwise.
        """
        path = safe_decode(path, enc=self.encoding)
        if os.path.exists(path) and not overwrite:
            return False
        from libopensesame.osexpfile import osexpwriter
        w = osexpwriter(self, path)
        if update_path:
            self.experiment_path = w.experiment_path
        return path

    def open(self, src):
        """Opens an experiment. The source can be any of the following:

        - An
        OpenSesame script (ie. not a file)
        - The full path to a .osexp
        experiment file.

        Parameters
        ----------
        src : str
            The source.

        Returns
        -------
        str
            An OpenSesame script.
        """
        from libopensesame.osexpfile import osexpreader
        f = osexpreader(self, src)
        self.experiment_path = f.experiment_path
        return f.script

    def reset_feedback(self):
        """Resets the feedback variables (acc, avg_rt, etc.)."""
        self.responses.reset_feedback()

    def var_info(self):
        """Returns a list of (name, value) tuples with variable descriptions
        for the main experiment.

        Returns:
        A list of tuples.
        """
        l = []
        for var in self.var:
            val = self.var.get(var, _eval=False)
            if self.var.is_default_loggable(var, val):
                l.append((var, val))
        return l

    def init_heartbeat(self):
        """Initializes heartbeat."""
        if self.heartbeat_interval <= 0 or self.var.fullscreen == 'yes' or \
                self.output_channel is None:
            self.heartbeat = None
            return
        from libopensesame.heartbeat import heartbeat
        self.heartbeat = heartbeat(self, interval=self.heartbeat_interval)
        self.heartbeat.start()

    def init_random(self):
        """Initializes the random number generators. For some reason, the numpy
        random seed is not re-initialized when the experiment is started
        again with the multiprocess runner, resulting in identical random
        runs. The standard random module doesn't suffer from this problem.
        But to be on the safe side, we now explicitly re-initialize the
        random seed.

        See also:

        - <http://forum.cogsci.nl/index.php?p=/discussion/1441/>
        """
        import random
        random.seed()
        try:
            # Don't assume that numpy is available
            import numpy
            numpy.random.seed()
        except:
            pass

    def init_sound(self):
        """Initializes the sound backend."""
        from openexp import sampler
        sampler.init_sound(self)

    def init_display(self):
        """Initializes the canvas backend."""
        from openexp import canvas
        canvas.init_display(self)
        self.python_workspace['win'] = self.window

    def init_clock(self):
        """Initializes the clock backend."""
        from openexp.clock import clock
        self._clock = clock(self)

    def init_log(self):
        """Initializes the log backend."""
        from openexp.log import log
        self._log = log(self, self.logfile)

# Alias for backwards compatibility
experiment = Experiment



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public Licensepass
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
from openexp import resources
from libopensesame.py3compat import *
from libopensesame import misc
from libopensesame.oslogging import oslogger
import warnings
from openexp._canvas._element.element import Element
from qtpy.QtWidgets import (
    QGraphicsTextItem,
    QStyleOptionGraphicsItem,
    QApplication
)
from qtpy.QtGui import QPixmap, QPainter, QColor, QFont, QFontDatabase, QImage
from qtpy.QtCore import Qt, QCoreApplication
from PIL import Image

FONTS = [
    'sans',
    'serif',
    'mono',
    'arabic',
    'chinese-japanese-korean',
    'hindi',
    'hebrew'
]
font_database = None
font_substitutions = []
pyqt_initialized = False


class RichText(Element):

    def __init__(
            self,
            canvas,
            text,
            center=True,
            x=None,
            y=None,
            max_width=None,
            **properties
    ):

        global pyqt_initialized

        if not pyqt_initialized:
            self._init_pyqt(canvas.experiment)
            pyqt_initialized = True
        x, y = canvas.none_to_center(x, y)
        properties = properties.copy()
        properties.update({
            'text': safe_decode(text),
            'center': center,
            'x': x,
            'y': y,
            'max_width': max_width
        })
        self._cached_size = None
        Element.__init__(self, canvas, **properties)

    def _init_pyqt(self, exp):

        global app, font_database

        # Add the Qt plugin folders to the library path, if they exists. Where
        # these folders are depends on the version of Qt4, but these are two
        # possible locations.
        qt_plugin_path = os.path.join(
            os.path.dirname(sys.executable), 'Library', 'plugins')
        if os.path.isdir(qt_plugin_path):
            QCoreApplication.addLibraryPath(
                safe_decode(qt_plugin_path, enc=sys.getfilesystemencoding())
            )
        qt_plugin_path = os.path.join(
            os.path.dirname(sys.executable), 'Library', 'lib', 'qt4', 'plugins')
        if os.path.isdir(qt_plugin_path):
            QCoreApplication.addLibraryPath(
                safe_decode(qt_plugin_path, enc=sys.getfilesystemencoding())
            )
        # If no instance of QApplication exists, a segmentation fault seems to
        # always occur. So we create one.
        if QCoreApplication.instance() is None:
            app = QApplication([])
        # Register the fonts bundled with OpenSesame
        if font_database is None:
            try:
                font_database = QFontDatabase()
            except TypeError:
                # PyQt6 the class is used directly, without creating an
                # instance
                font_database = QFontDatabase
            for font in FONTS:
                self._register_font(exp, font)

    def _register_font(self, exp, font, fd=None):

        global pyqt_initialized
        if not pyqt_initialized:
            self._init_pyqt(exp)
            pyqt_initialized = True
        oslogger.debug(u'Registering font {}'.format(font))
        try:
            path = resources[f'{font}.ttf']
        except Exception:
            warnings.warn(u'Font %s not found' % font)
        else:
            if fd is None:
                fd = font_database
            font_id = fd.addApplicationFont(str(path))
            if font_id < 0:
                warnings.warn(u'Failed to load font %s' % font)
                return
            font_families = fd.applicationFontFamilies(font_id)
            if not font_families:
                warnings.warn(u'Font %s contains no families' % font)
                return
            font_substitutions.append((font_families[0], font))

    def _register_custom_font(self, font):

        if font + u'.ttf' not in self.experiment.pool:
            return
        if font in (font for font_family, font in font_substitutions):
            return
        self._register_font(self.experiment, font)

    @property
    def size(self):

        if self._cached_size:
            return self._cached_size
        bbox = self._qimage_to_pil(self._to_qimage()).getbbox()
        x1, y1, x2, y2 = (0, 0, 1, 1) if bbox is None else bbox
        y2 = max(y1 + self.font_size, y2)
        self._cached_size = x2 - x1, y2 - y1
        return self._cached_size

    @property
    def rect(self):

        w, h = self.size
        if self.center:
            return self.x - w // 2, self.y - h // 2, w, h
        return self.x, self.y, w, h

    def _to_qgraphicstextitem(self):

        t = QGraphicsTextItem()
        t.setDefaultTextColor(QColor(self.color.hexcolor))
        if self.html:
            text = self.text.replace(u'\n', u'<br />')
            t.setHtml(
                u'<div align="center">%s</div>' % text
                if self.center else text
            )
        else:
            t.setPlainText(self.text)
        mw = self.max_width
        if mw is None:
            mw = self._canvas.width // 2 - self.x
        if self.center:
            mw *= 2
        t.setTextWidth(mw)
        # Register custom fonts that are placed in the file pool
        self._register_custom_font(self.font_family)
        f = QFont(
            self.font_family,
            weight=QFont.Bold if self.font_bold else QFont.Normal,
            italic=self.font_italic
        )
        for family, substitute in font_substitutions:
            f.insertSubstitution(substitute, family)
        f.setPixelSize(self.font_size)
        t.setFont(f)
        return t

    def _to_qimage(self):

        t = self._to_qgraphicstextitem()
        rect = t.boundingRect()
        height = int(rect.height())
        width = int(rect.width())
        pixmap = QPixmap(width, height)
        pixmap.fill(Qt.transparent)
        painter = QPainter(pixmap)
        painter.setPen(Qt.red)
        if self.center and self.html:
            painter.drawPoint(width // 2, 0)
            painter.drawPoint(width // 2, height - 1)
        else:
            painter.drawPoint(0, 0)
            painter.drawPoint(0, height - 1)
        t.paint(painter, QStyleOptionGraphicsItem(), None)
        painter.end()
        return pixmap.toImage()
        
    def _qimage_to_pil(self, qimage):
        """This functions takes a QImage and returns a PIL Image. If available,
        this uses PIL's built-in conversion tool, but since this has broken
        for PyQt5 for Pillow >= 10, we resort to a custom conversion procedure
        if necessary.
        """
        try:
            return Image.fromqimage(self._to_qimage())
        except (ImportError, TypeError):
            pass
        import numpy as np
        oslogger.info('no Qt bindings, using custom QImage to PIL conversion')
        qimage = qimage.convertToFormat(QImage.Format_ARGB32)
        width = qimage.width()
        height = qimage.height()
        ptr = qimage.bits()
        ptr.setsize(qimage.byteCount())
        # Assuming 32-bit color depth
        arr = np.array(ptr).reshape(height, width, 4)
        if arr.shape[2] == 4:
            arr = arr[..., [2, 1, 0, 3]]  # Convert ARGB to RGBA
        else:
            arr = arr[..., [2, 1, 0]]  # Convert RGB to BGR
        return Image.fromarray(arr)

    def _to_pil(self):

        im = self._qimage_to_pil(self._to_qimage())
        bbox = im.getbbox()
        x1, y1, x2, y2 = (0, 0, 1, 1) if bbox is None else bbox
        y1 = min(y2 - self.font_size, y1)
        return im.crop((x1, y1 + 1, x2, y2 - 1))

    @staticmethod
    def _setter(key, self, val):

        if key == u'text':
            val = safe_decode(val)
        super(RichText, self)._setter(key, self, val)

    def _on_attribute_change(self, **kwargs):

        self._cached_size = None
        super(RichText, self)._on_attribute_change(key, self, val)



"""A simple text display form"""

category = "Form"
icon = "os-form_text_display"
help_url = "manual/forms/readymade"
controls = [
    {
        "type": "line_edit",
        "var": "form_title",
        "label": "Form title",
        "tooltip": "Title to appear above the form text"
    },
    {
        "type": "line_edit",
        "var": "ok_text",
        "label": "Ok-button text",
        "tooltip": "Text for the Ok button"
    },
    {
        "type": "editor",
        "var": "form_text",
        "label": "Main form text",
        "tooltip": "Text to display in the form body"
    }
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import time
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension
from libopensesame.oslogging import oslogger
from qtpy.QtCore import QAbstractEventDispatcher, QTime


class PreloadItems(BaseExtension):

    def event_startup(self):

        self._monitoring_idle = False
        self._interval = cfg.preload_items_interval
        self._threshold = cfg.preload_items_idle_threshold
        self._start_idle_monitor()

    def _start_idle_monitor(self):

        if self._monitoring_idle:
            return
        self._idle = False
        self._monitoring_idle = True
        self._aed = QAbstractEventDispatcher.instance()
        self._aed.awake.connect(self._on_awake)
        self._n_events = 0
        self._last_update = QTime.currentTime()
        oslogger.debug('starting idle monitor')

    def _stop_idle_monitor(self):

        if not self._monitoring_idle:
            return
        self._aed.awake.disconnect()
        self._monitoring_idle = False
        oslogger.debug('stopping idle monitor')

    def _on_awake(self):

        self._n_events += 1
        self._last_awake = QTime.currentTime()
        time_past = self._last_update.msecsTo(self._last_awake)
        if time_past < self._interval:
            return
        event_rate = self._n_events / time_past
        if event_rate < self._threshold:
            self._idle = True
            oslogger.debug('idle {:4f}'.format(event_rate))
            self._preload_one_item()
        else:
            self._idle = False
            oslogger.debug('busy {:4f}'.format(event_rate))
        self._last_update = self._last_awake
        self._n_events = 0

    def _preload_one_item(self):

        for name, item in self.item_store.items():
            if item.container_widget is not None:
                continue
            t = time.time()
            self.extension_manager.fire(u'notify_suspend')
            item.init_edit_widget()
            item.edit_widget()
            item.first_refresh = True
            self.extension_manager.fire(u'notify_resume')
            oslogger.debug('preloaded {} in {:.2f} ms\n'.format(
                name,
                1000 * (time.time() - t))
            )
            break
        else:
            self._stop_idle_monitor()

    def event_run_experiment(self, fullscreen):

        self._stop_idle_monitor()

    def event_end_experiment(self, ret_val):

        self._start_idle_monitor()

    def event_regenerate(self):

        self._start_idle_monitor()

    def event_open_experiment(self, path):

        self._start_idle_monitor()



"""Periodically saves your experiment to a back-up folder."""

icon = "folder-open"
label = "Open backup folder"
menu = {
    "index": 1,
    "separator_after": True,
    "submenu": "Tools"
}
settings = {
    "autosave_interval": 600000,
    "autosave_max_age": 7
}



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from qtpy import QtCore, QtWidgets, uic
from openexp import resources
from libopensesame.oslogging import oslogger


class BaseComponent:

    r"""A base class for all components, notably dialogs and widgets."""
    enc = u'utf-8'

    def setup(self, main_window, ui=None):
        r"""Constructor.

        Parameters
        ----------
        main_window
            A qtopensesame object.
        ui, optional
            An id for a user-interface file. For example
            'dialogs.quick_switcher' will correspond to the file
            'resources/ui/dialogs/quick_switcher.ui'.
        """
        self.main_window = self.get_main_window(main_window)
        self.load_ui(ui)
        if hasattr(self.main_window, u'theme'):
            self.main_window.theme.apply_theme(self)

    def load_ui(self, ui=None):
        r"""Dynamically loads the ui, if any.

        Parameters
        ----------
        ui, optional
            An id for a user-interface file, or None.
        """
        self.ui = None
        if ui is None:
            return
        # If the UI file has been explicitly registered, which is the case
        # for extensions
        try:
            ui_path = resources[ui]
        except FileNotFoundError:
            # Dot-split the ui id, append a `.ui` extension, and assume it's
            # relative to the resources/ui subfolder.
            path_list = [u'ui'] + ui.split(u'.')
            try:
                ui_path = resources[os.path.join(*path_list) + '.ui']
            except FileNotFoundError:
                return
        oslogger.debug(f'dynamically loading ui: {ui_path}')
        with safe_open(ui_path) as fd:
            self.ui = uic.loadUi(fd, self)

    def get_main_window(self, main_window):
        r"""If the main_window is actually not the main window, but a widget
        that has the main window somewhere above it in the hierarchy, we
        traverse upwards.

        Parameters
        ----------
        main_window
            An object that is the main window or a descendant of the main
            window.

        Returns
        -------
        qtopensesame
            The main window.
        """
        from libqtopensesame.qtopensesame import QtOpenSesame
        while not isinstance(main_window, QtOpenSesame):
            if hasattr(main_window, u'main_window'):
                _parent = main_window.main_window
            else:
                _parent = main_window.parent()
            main_window = _parent
        return main_window

    @staticmethod
    def quick_connect(slot, signals):
        r"""A convenience function to connect many signals to one slot.

        Parameters
        ----------
        slot
            a slot
        signals
            a list of signals
        """
        for signal in signals:
            signal.connect(slot)


# Alias for backwards compatibility
base_component = BaseComponent



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp.canvas import Canvas
from ..form_base.form_base import FormBase

default_script = u"""
__form_text__
You are about to participate in an experiment.

This experiment poses no known risks to your health and your name will not be associated with the findings.

Upon completion of your participation in this study you will be provided with a brief explanation of the question this study addresses.

If you have any questions not addressed by this consent form, please do not hesitate to ask.

You can stop at any time during experiment if you feel uncomfortable.
__end__
set form_title 'Consent form'
set checkbox_text 'I have read and understood the information shown above'
set accept_text 'Participate!'
set decline_text 'Do not participate'
set decline_message 'You need to accept the consent form to participate!'
set rows 1;4;1;1
widget 0 0 2 1 label text={form_title}
widget 0 1 2 1 label text={form_text} center=no
widget 0 2 2 1 checkbox text={checkbox_text} var=checkbox_status
widget 0 3 1 1 button text={accept_text} var=accept_status
widget 1 3 1 1 button text={decline_text}
"""


class FormConsent(FormBase):

    initial_view = u'controls'

    def __init__(self, name, experiment, string=None):
        if string is None or string.strip() == u'':
            string = default_script
        super().__init__(name, experiment, string)

    def run(self):
        while True:
            super().run()
            if self.var.get(u'checkbox_status') == \
                    self.var.get(u'checkbox_text') and \
                    self.var.get(u'accept_status') == u'yes':
                break
            c = Canvas(self.experiment)
            c.text(self.var.get(u'decline_message'))
            c.show()
            self.clock.sleep(5000)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import backend


class WidgetFactory:

    def __init__(self, *args, **kwargs):

        self._args = args
        self._kwargs = kwargs

    def construct(self, form):

        mod = __import__(
            'libopensesame.widgets._%s' % self.mod,
            fromlist=['dummy']
        )
        cls = getattr(mod, self.__class__.__name__)
        return cls(form, *self._args, **self._kwargs)


class Label(WidgetFactory):
    mod = 'label'


class Button(WidgetFactory):
    mod = 'button'


class ImageButton(WidgetFactory):
    mod = 'image_button'


class ImageWidget(WidgetFactory):
    mod = 'image'


class RatingScale(WidgetFactory):
    mod = 'rating_scale'


class TextInput(WidgetFactory):
    mod = 'text_input'


class Checkbox(WidgetFactory):
    mod = 'checkbox'



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import json
import time
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import osexception
from libopensesame.py3compat import *

# Caching variables
_plugin_dict = {}
_folders = {}
_properties = {}

# The plug-ins can be either source or bytecode. Usually they will be source,
# but some distributions (notably the runtime for Android) will automatically
# compile everything to bytecode.
src_templates = [u'%s.py']
bytecode_templates = [u'%s.pyo', u'%s.pyc']


def plugin_folders(only_existing=True, _type=u'plugins'):
    r"""Returns a list of plugin folders.

    Parameters
    ----------
    only_existing : bool, optional
        Specifies if only existing folders should be returned.
    _type : str, unicode, optional
        Indicates whether runtime plugins ('plugins') or GUI extensions
        ('extensions') should be listed.

    Returns
    -------
    list
        A list of folders.
    """
    l = []
    # Build a list of default plugin/ extension folders
    for folder in misc.base_folders:
        path = os.path.join(folder, u'opensesame_%s' % _type)
        if os.path.exists(path):
            l.append(path)
    # Get the environment variables
    if _type == u'plugins':
        key = u'OPENSESAME_PLUGIN_PATH'
    else:
        key = u'OPENSESAME_EXTENSION_PATH'
    if key in os.environ:
        for path in os.environ[key].split(';'):
            path = safe_decode(path, enc=sys.getfilesystemencoding())
            if os.path.exists(path):
                l.insert(0, path)
    if os.name == u'posix' and u'HOME' in os.environ:
        # Regular Linux distributions. TODO: How well does this apply to Mac OS?
        path = os.path.join(os.environ[u'HOME'], u'.opensesame', _type)
        if not only_existing or os.path.exists(path):
            l.append(path)
        path = u'/usr/share/opensesame/%s' % _type
        if not only_existing or os.path.exists(path):
            l.append(path)
    elif os.name == u'posix' and u'HOME' not in os.environ:
        # Android can be recognized by the fact that the HOME variable is not
        # available. We can simply use the relative path `plugins`, which will
        # (always?) point to `/data/data/nl.cogsci.nl/opensesame/files/plugins`
        path = _type
        if not only_existing or os.path.exists(path):
            l.append(path)
    elif os.name == u'nt':
        # Windows
        path = os.path.join(safe_decode(os.environ[u'APPDATA'],
                                        enc=sys.getfilesystemencoding()),
                                        u'.opensesame', _type)
        if not only_existing or os.path.exists(path):
            l.append(path)
    return l


def is_plugin(item_type, _type=u'plugins'):
    r"""Checks if a given item type corresponds to a plugin.

    Returns
    -------
    True if the item_type is a plugin, False otherwise.
    """
    return plugin_folder(item_type, _type=_type) is not None


def plugin_disabled(plugin, _type=u'plugins'):
    r"""Checks if a plugin has been disabled. If the config module cannot be
    loaded the return value is False.

    Parameters
    ----------
    plugin
        The plugin to check.
    _type, optional
        plugins/ extensions

    Returns
    -------
    True if the plugin has been disabled, False otherwise.
    """
    from libqtopensesame.misc.config import cfg

    if plugin_property(plugin, u'disabled', _type=_type):
        return True
    # Check if disabled_plugins is not empty in which case it is a
    # QPyNullVariant which does not have split() method
    disabled_plugins = cfg[u'disabled_%s' % _type]
    if hasattr(disabled_plugins, u"split") and plugin in disabled_plugins.split(u';'):
        return True
    return False


def plugin_property(plugin, _property, default=0, _type=u'plugins'):
    r"""Returns a property of a plug-in.

    Parameters
    ----------
    plugin
        The name of the plugin.
    _property
        The name of the property.
    default, optional
        A default property value.

    Returns
    -------
    The property value.
    """
    return plugin_properties(plugin, _type=_type).get(_property, default)


def set_plugin_property(plugin, property, value):
    r"""Sets a property in the info dictionary for a plugin.

    Parameters
    ----------
    plugin
        The plugin name.
    property
        The property name.
    value
        The property value.
    """
    plugin_properties(plugin)
    _properties[plugin][property] = value


def plugin_properties(plugin, _type=u'plugins'):
    r"""Gets the info dictionary for a plugin.

    Parameters
    ----------
    plugin
        The plugin name.
    _type
        The plugin type (i.e. 'plugins' or 'extensions').

    Returns
    -------
    An info dictionary.
    """
    if plugin in _properties:
        return _properties[plugin]
    folder = plugin_folder(plugin, _type=_type)
    info_txt = os.path.join(folder, u'info.txt')
    info_json = os.path.join(folder, u'info.json')
    info_yaml = os.path.join(folder, u'info.yaml')
    # First try to read json. This is used by the plugins bundled with
    # OpenSesame because it is faster (even if it's less readable)
    if os.path.exists(info_json):
        with safe_open(info_json) as fd:
            s = fd.read()
        try:
            _properties[plugin] = json.loads(s)
        except:
            oslogger.error(u'Failed to parse %s' % info_json)
            _properties[plugin] = {}
    # New-style plug-ins, using info.yaml
    elif os.path.exists(info_yaml):
        # Read the yaml file and replace all tabs by spaces. This is necessary,
        # because yaml doesn't accept tab-based indentation, whereas json does.
        with safe_open(info_yaml) as fd:
            s = fd.read()
        s = s.replace(u'\t', u'    ')
        try:
            _properties[plugin] = safe_yaml_load(s)
        except:
            oslogger.error(u'Failed to parse %s' % info_yaml)
            _properties[plugin] = {}
    # Old-style plug-ins, using info.txt
    elif os.path.exists(info_txt):
        _properties[plugin] = {}
        with safe_open(info_txt) as fd:
            for l in fd:
                a = l.split(":")
                if len(a) == 2:
                    val = a[1].strip()
                    var = a[0].strip()
                    try:
                        val = int(val)
                    except:
                        pass
                    _properties[plugin][var] = val
    else:
        _properties[plugin] = {}
        oslogger.error(
            u'Failed to read plug-in information (%s) from info.[txt|yaml|json]'
            % plugin
        )
    _properties[plugin][u'plugin_folder'] = folder
    _properties[plugin][u'type'] = _type
    return _properties[plugin]


def plugin_category(plugin, _type=u'plugins'):
    r"""Returns the category of a plugin.

    Parameters
    ----------
    plugin
        The plugin name.
    _type, optional
        plugins/ extensions

    Returns
    -------
    A category
    """
    return plugin_property(
        plugin,
        u'category',
        default=u'Miscellaneous',
        _type=_type
    )


def list_plugins(filter_disabled=True, _type=u'plugins', mode=u'default'):
    r"""Returns a list of plugins.

    Parameters
    ----------
    filter_disabled, optional
        Indicates whether disabled plugins should be included in the list.
    _type, optional
        plugins/ extensions
    mode, optional
        A mode as specified on the command line

    Returns
    -------
    A list of plugins (item_types).
    """
    if _type not in _plugin_dict:
        plugins = []
        for folder in plugin_folders(_type=_type):
            for plugin in os.listdir(folder):
                if is_plugin(plugin, _type=_type):
                    _plugin = plugin, plugin_property(
                        plugin, u'priority',
                        _type=_type
                    )
                    if _plugin not in plugins:
                        plugins.append(_plugin)
        # Sort (inversely) by priority
        plugins.sort(key=lambda p: -p[1])
        _plugin_dict[_type] = [plugin[0] for plugin in plugins]
    return [
        plugin
        for plugin in _plugin_dict[_type]
        if not (filter_disabled and plugin_disabled(plugin, _type=_type))
    ]


def plugin_folder(plugin, _type=u'plugins'):
    r"""Returns the folder of a plugin

    Parameters
    ----------
    plugin
        The name of the plugin

    Returns
    -------
    The folder of the plugin.
    """
    global _folders

    if plugin in _folders:
        return _folders[plugin]
    for folder in plugin_folders(_type=_type):
        plugin = str(plugin)
        for tmpl in src_templates + bytecode_templates:
            if os.path.exists(os.path.join(folder, plugin, tmpl % plugin)):
                f = os.path.join(folder, plugin)
                _folders[plugin] = f
                return f
    return None


def plugin_icon_large(plugin, _type=u'plugins'):
    r"""Return the large icon for a plugin

    Parameters
    ----------
    plugin
        The name of the plugin
    _type, optional
        plugins/ extensions

    Returns
    -------
    The full path to an icon
    """
    icon = plugin_property(plugin, u'icon', default=None)
    if icon is not None:
        return icon
    return os.path.join(plugin_folder(plugin, _type=_type),
                        u'%s_large.png' % plugin)


def plugin_icon_small(plugin, _type=u'plugins'):
    r"""Return the small icon for a plugin

    Parameters
    ----------
    plugin
        The name of the plugin.
    _type, optional
        plugins/ extensions

    Returns
    -------
    The full path to an icon
    """
    icon = plugin_property(plugin, u'icon', default=None)
    if icon is not None:
        return icon
    return os.path.join(plugin_folder(plugin, _type=_type), u'%s.png' % plugin)


def import_plugin(plugin, _type=u'plugins'):
    """Imports plugin module.

    Parameters
    ----------
    plugin
        The name of the plugin.
    _type, optional
        plugins/ extensions

    Returns
    -------
    The imported module.
    """
    import importlib.util
    
    plugin = str(plugin)
    folder = plugin_folder(plugin, _type=_type)
    for tmpl in src_templates:
        if os.path.exists(os.path.join(folder, tmpl % plugin)):
            path = os.path.join(folder, tmpl % plugin)
            # Create a module spec from the source file and load the module
            spec = importlib.util.spec_from_file_location(plugin, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
    for tmpl in bytecode_templates:
        if os.path.exists(os.path.join(folder, tmpl % plugin)):
            path = os.path.join(folder, tmpl % plugin)
            # Similarly load the module from compiled bytecode
            spec = importlib.util.spec_from_file_location(plugin, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module


def load_plugin(
        plugin, item_name, experiment, string, prefix=u'', _type=u'plugins'
):
    r"""Returns an instance of the plugin.

    Parameters
    ----------
    plugin
        The name of the plugin.
    item_name
        The name of the item (plugin instance).
    experiment
        The experiment object.
    string
        A definition string.
    prefix, optional
        A class prefix to allow switching between [plugin] and qt[plugin].

    Returns
    -------
    An item (plugin instance).
    """
    t0 = time.time()
    if plugin_folder(plugin, _type=_type) not in sys.path:
        sys.path.append(plugin_folder(plugin, _type=_type))
    item_module = import_plugin(plugin, _type=_type)
    item_class = getattr(item_module, prefix + plugin)
    item = item_class(item_name, experiment, string)
    set_plugin_property(plugin, u'startup_time', time.time() - t0)
    return item


def load_extension(ext_name, main_window):
    r"""Creates an instance of an extension.

    Parameters
    ----------
    ext_name
        The extension name.
    main_window
        The main window object.

    Returns
    -------
    An extension object.
    """
    t0 = time.time()
    sys.path.append(plugin_folder(ext_name, _type=u'extensions'))
    mod = import_plugin(ext_name, _type=u'extensions')
    cls = getattr(mod, ext_name)
    ext = cls(main_window)
    set_plugin_property(ext_name, u'startup_time', time.time() - t0)
    return ext


def load_cls(path, cls, mod, pkg=None):
    r"""Dynamically loads a module from a path and return a class from the
    module.

    Parameters
    ----------
    path
        A folder or file name. If a filename is specified, the file's directory
        will be used as path.
    cls
        The class name.
    mod
        The name of a module, which corresponds to the name of the source file
        without the `.py` extension.
    pkg
        The module's package. This is effectively a subfolder that is added to
        the path.

    Returns
    -------
    A class.
    """
    mod = load_mod(path, mod, pkg)
    return getattr(mod, cls)


def load_mod(path, mod, pkg=None):
    r"""Dynamically loads a module from a path.

    Parameters
    ----------
    path
        A folder or file name. If a filename is specified, the file's directory
        will be used as path.
    mod
        The name of a module, which corresponds to the name of the source file
        without the `.py` extension.
    pkg
        The module's package. This is effectively a subfolder that is added to
        the path.

    Returns
    -------
    A module.
    """
    import importlib.util
    
    path = safe_decode(path, enc=sys.getfilesystemencoding())
    if not os.path.isdir(path):
        path = os.path.dirname(path)
    if pkg is not None:
        path = os.path.join(path, pkg)
    path = os.path.join(path, mod + '.py')
    if not os.path.exists(path):
        raise osexception('%s does not exist' % path)
    oslogger.debug('loading module from %s' % path)
    # Create a module spec from the source file and load the module
    spec = importlib.util.spec_from_file_location(mod, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from libopensesame.item import Item
import random


class AdvancedDelay(Item):

    def reset(self):
        self.var.duration = 1000
        self.var.jitter = 0
        self.var.jitter_mode = u'Uniform'

    def prepare(self):
        super().prepare()
        if not isinstance(self.var.duration, (int, float)) \
                or self.var.duration < 0:
            raise InvalidValue('Duration should be a positive numeric value')
        if self.var.jitter_mode == u'Uniform':
            self._duration = random.uniform(
                self.var.duration-self.var.jitter / 2,
                self.var.duration+self.var.jitter / 2)
        elif self.var.jitter_mode == u'Std. Dev.':
            self._duration = random.gauss(self.var.duration, self.var.jitter)
        else:
            raise InvalidValue(f'Unknown jitter mode: {self.var.jitter_mode}')
        if self._duration < 0:
            self._duration = 0
        self._duration = int(self._duration)
        self.experiment.var.set(f'delay_{self.name}', self._duration)

    def run(self):
        self.set_item_onset(self.clock.time())
        self.clock.sleep(self._duration)

    def var_info(self):
        return super().var_info() + \
            [(f'delay_{self.name}', '[Determined at runtime]')]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from libqtopensesame.items.sequence import Sequence
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.items.qtstructure_item import QtStructureItem
from libopensesame.sequence import Sequence as SequenceRuntime
from libqtopensesame.widgets.tree_item_item import TreeItemItem
from libqtopensesame.validators import DurationValidator
from . import Coroutines, ItemsAdapter, TreeOverviewAdapter
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'coroutines', category=u'plugins')


class QtCoroutines(Coroutines, Sequence):

    help_url = u'manual/structure/coroutines'

    def __init__(self, name, experiment, string=None):
        Coroutines.__init__(self, name, experiment, string)
        # We don't call the sequence constructor, because it doesn't specify
        # the plugin_file to qtplugin, which we need to do. We therefore
        # explicitly need to call the qtplugin and qtstructure_item
        # constructors.
        SequenceRuntime.__init__(self, name, experiment, string)
        QtStructureItem.__init__(self)
        QtPlugin.__init__(self, plugin_file=os.path.dirname(__file__))
        self.last_removed_child = None, None

    def reset(self):
        Coroutines.reset(self)
        # Recreate the items adapter when the schedule is re-initialized.
        self._items = ItemsAdapter(self.schedule)

    @property
    def items(self):
        r"""A property that maps the schedule list to an items list as expected
        by sequence.
        """
        return self._items

    @items.setter
    def items(self, val):
        r"""A setter that maps an items list to a schedule list."""
        self.schedule = self.schedule[:len(val)]
        self._items = ItemsAdapter(self.schedule)
        for i, (item_name, cond, enabled) in enumerate(val):
            start_time = self.schedule[i][1]
            end_time = self.schedule[i][2]
            self.schedule[i] = item_name, start_time, end_time, cond, enabled

    def init_edit_widget(self):
        super(Sequence, self).init_edit_widget(stretch=False)
        self.treewidget = TreeOverviewAdapter(self, self.main_window,
                                              overview_mode=False)
        self.treewidget.setup(self.main_window)
        self.treewidget.structure_change.connect(self.update)
        self.treewidget.text_change.connect(self.update_script)
        self.treewidget.setHeaderLabels([_(u'Item name'), (u'Run if'),
                                         _(u'Start time'),
                                         _(u'End time (if applicable)')])
        self.set_focus_widget(self.treewidget)
        self.edit_vbox.addWidget(self.treewidget)
        self.add_line_edit_control(u'duration', _(u'Duration'),
            validator=DurationValidator(self.main_window, default=u'5000'))
        self._combobox_end_after_item = self.add_combobox_control(
            u'end_after_item', _(u'End after item (optional)'),
            options=[s[0] for s in self.schedule]
        )
        self.add_checkbox_control(u'flush_keyboard',
                                  _(u'Flush pending key presses at coroutines start'))

    def edit_widget(self):
        super().edit_widget()
        self._refresh()

    def rename(self, from_name, to_name):
        super().rename(from_name, to_name)
        # The end-after-item has changed we need to refresh and also update
        # corersponding variable
        if self.var.end_after_item == from_name:
            self.var.end_after_item = to_name
            self._refresh()
        else:
            # If any other items have changed, we just need to refrsh. We
            # compare to the to_name, because the renaming has already happened
            # at this point through the super().rename()
            for name, *_ in self._items:
                if name == to_name:
                    self._refresh()

    def build_item_tree(self, toplevel=None, items=[], max_depth=-1,
                        extra_info=None):
        widget = TreeItemItem(self, extra_info=extra_info)
        items.append(self.name)
        if max_depth < 0 or max_depth > 1:
            for item, start_time, end_time, cond, enabled in self.schedule:
                if item in self.experiment.items:
                    self.experiment.items[item].build_item_tree(
                        widget, items, max_depth=max_depth-1, extra_info=cond)
                    child = widget.child(widget.childCount()-1)
                    if enabled:
                        child.setText(2, safe_decode(start_time))
                        if not self.is_oneshot_coroutine(item):
                            child.setText(3, safe_decode(end_time))
                    else:
                        child.setDisabled(True)
                        child.setText(1, None)
                        child.setText(2, None)
                        child.setText(3, None)
        if toplevel is not None:
            toplevel.addChild(widget)
        else:
            widget.set_draggable(False)
        return widget

    def insert_child_item(self, item_name, index=0):
        if not self.is_coroutine(item_name):
            self.extension_manager.fire(
                'notify',
                message=_('"%s" does not support coroutines.') % item_name)
            return
        super().insert_child_item(item_name, index=index)
        self._refresh()

    def remove_child_item(self, item_name, index=0):
        super().remove_child_item(item_name, index=index)
        if item_name == self.var.end_after_item:
            self.var.end_after_item = u''
        self._refresh()

    def _refresh(self):
        r"""Refreshes the GUI, which currently means updating the combobox for
        end_after_item.
        """
        # No refresh necessary if the combobox hasn't been initialized yet
        if not hasattr(self, u'_combobox_end_after_item'):
            return
        self._combobox_end_after_item.clear()
        self._combobox_end_after_item.addItem(u'')
        current_index = 0
        for i, (item_name, st, et, cond, enabled) in enumerate(self.schedule):
            if self.var.end_after_item == item_name:
                current_index = i+1
            self._combobox_end_after_item.addItem(
                self.theme.qicon(self.experiment.items[item_name].item_icon()),
                item_name)
        self._combobox_end_after_item.setCurrentIndex(current_index)



"""Command palette"""

tooltip = "Trigger menu actions"
label = "Command palette"
shortcut = "Ctrl+Shift+P"
icon = "edit-find"
priority = -2
menu = {
    "index": 3,
    "submenu": "View"
}
modes = ["default", "ide"]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._coordinates.psycho import Psycho as PsychoCoordinates
from openexp._mouse.mouse import Mouse
from psychopy import event
from openexp.backend import configurable


class Psycho(Mouse, PsychoCoordinates):

    r"""This is a mouse backend built on top of PsychoPy. For function
    specifications and docstrings, see `openexp._mouse.mouse`.
    """
    def __init__(self, experiment, **resp_args):

        Mouse.__init__(self, experiment, **resp_args)
        PsychoCoordinates.__init__(self)
        self.mouse = event.Mouse(visible=False, win=self.experiment.window)
        event.mouseButtons = [0, 0, 0]

    @configurable
    def get_click(self):

        buttonlist = self.buttonlist
        timeout = self.timeout
        self.mouse.setVisible(self.visible)
        start_time = self.experiment.clock.time()
        button = None
        pos = None
        self.mouse.clickReset()
        while True:
            time = self.experiment.clock.time()
            buttons, times = self.mouse.getPressed(getTime=True)
            for i in (1, 2, 3):
                if buttons[i-1] and (buttonlist is None or i in buttonlist):
                    button = i
                    pos = self.mouse.getPos()
                    break
            else:
                if timeout is not None and time-start_time >= timeout:
                    break
                continue
            break
        if pos is not None:
            pos = self.from_xy(pos)
        self.mouse.setVisible(self._cursor_shown)
        return button, pos, time

    def show_cursor(self, show=True):

        self.mouse.setVisible(show)
        Mouse.show_cursor(self, show=show)

    def get_pos(self):

        return self.from_xy(self.mouse.getPos()), self.experiment.clock.time()

    def set_pos(self, pos=(0, 0)):

        self.mouse.setPos(self.to_xy(pos))

    def get_pressed(self):

        return tuple(self.mouse.getPressed(getTime=False))

    def flush(self):

        event.mouseButtons = [0, 0, 0]
        event.clearEvents()
        return False


# Non PEP-8 alias for backwards compatibility
psycho = Psycho


# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy.QtWidgets import QTabBar, QMenu, QAction
from qtpy.QtCore import Qt
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tab_bar', category=u'core')


class TabBar(BaseSubcomponent, QTabBar):
    """A tab-bar for the overview area, implementing a right-click context
    menu, and a close-on-middle-click.
    """
    def __init__(self, tab_widget):

        QTabBar.__init__(self, tab_widget)
        BaseSubcomponent.setup(self, tab_widget)
        self._tab_widget = tab_widget
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)

    def mousePressEvent(self, event):

        QTabBar.mousePressEvent(self, event)
        if event.button() == Qt.MiddleButton:
            self._tab_widget.removeTab(self.tabAt(event.pos()))

    def _show_context_menu(self, pos):

        tab_index = self.tabAt(pos)
        self._tab_widget.setCurrentIndex(tab_index)
        menu = QMenu(self)
        menu.addAction(self.main_window.ui.action_close_current_tab)
        menu.addAction(self.main_window.ui.action_close_all_tabs)
        menu.addAction(self.main_window.ui.action_close_other_tabs)
        if 'tab_to_dockwidget' in self.extension_manager:
            menu.addSeparator()
            menu.addAction(
                self.extension_manager['tab_to_dockwidget'].action
            )
        menu.exec_(self.mapToGlobal(pos))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from libopensesame.oslogging import oslogger
from libopensesame.item import Item
try:
    import Quest
except ImportError:
    oslogger.debug('Failed to import Quest, trying psychopy.contrib.quest')
    from psychopy.contrib import quest as Quest


class QuestStaircaseInit(Item):

    def reset(self):
        self.var.t_guess = .5
        self.var.t_guess_sd = .25
        self.var.p_threshold = .75
        self.var.beta = 3.5
        self.var.delta = .01
        self.var.gamma = .5
        self.var.quest_name = 'default'
        self.var.test_value_method = 'quantile'
        self.var.min_test_value = 0
        self.var.max_test_value = 1
        self.var.var_test_value = 'quest_test_value'

    def quest_set_next_test_value(self):
        quest = self.experiment.quest[self.var.quest_name]
        if self.var.test_value_method == 'quantile':
            fnc = quest.quantile
        elif self.var.test_value_method == 'mean':
            fnc = quest.mean
        elif self.var.test_value_method == 'mode':
            fnc = quest.mode
        else:
            raise InvalidValue(
                f'Unknown test_value_method {self.var.test_value_method}')
        test_value = max(self.var.min_test_value,
                         min(self.var.max_test_value, fnc()))
        oslogger.debug(f'quest_test_value = {test_value}')
        self.experiment.var.quest_test_value = test_value
        self.experiment.var.set(self.var.var_test_value, test_value)

    def prepare(self):
        if not hasattr(self.experiment, 'quest'):
            self.experiment.quest = {}
            self.experiment.quest_set_next_test_value = {}
        self.experiment.quest[self.var.quest_name] = Quest.QuestObject(
            self.var.t_guess, self.var.t_guess_sd, self.var.p_threshold,
            self.var.beta, self.var.delta, self.var.gamma)
        self.experiment.quest_set_next_test_value[self.var.quest_name] = \
            self.quest_set_next_test_value
        self.quest_set_next_test_value()

    def var_info(self):
        """Gives a list of dictionaries with variable descriptions.

        Returns
        -------
        list
            A list of (name, description) tuples.
        """
        return super().var_info() + [('quest_test_value',
                                      '(Determined by Quest procedure)')]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.mouse_response import MouseResponse


class TouchResponse(MouseResponse):

    def reset(self):

        self.var._ncol = 2
        self.var._nrow = 1
        super().reset()

    def process_response(self, response_args):
        response, pos, t1 = response_args
        if pos is not None:
            x, y = pos
            _x = x+self.experiment.var.width/2
            _y = y+self.experiment.var.height/2
            col = _x // (self.experiment.var.width / self.var._ncol)
            row = _y // (self.experiment.var.height / self.var._nrow)
            response = row * self.var._ncol + col + 1
        super().process_response((response, pos, t1))



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import uuid
import os
import ast
import time
import json
import inspect
import pickle
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy.QtWidgets import QApplication
from qtpy.QtCore import QTimer
from libopensesame.oslogging import oslogger
from qtconsole.rich_jupyter_widget import RichJupyterWidget
try:
    from base64 import decodebytes
except ImportError:  # Python 2
    from base64 import decodestring as decodebytes
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'JupyterConsole', category=u'extension')


GLOBAL_EXPR = u'''{
key: (
    repr(val),
    val.__class__.__name__,
    (
        val.shape if hasattr(val, 'shape')
        else len(val) if hasattr(val, '__len__')
        else len(val.layers) if hasattr(val, 'layers') and hasattr(val.layers, '__len__')
        else None
    )
)
for key, val in globals().items()
if (
    hasattr(val, 'layers') and hasattr(val.layers, '__len__')
) or (
    not key.startswith(u'_') and
    key not in ('In', 'Out') and
    not callable(val) and
    not inspect.isclass(val) and
    not inspect.ismodule(val)
)
}'''

LIST_GLOBAL_EXPR = u'''[
key
for key, val in globals().items()
if (
    hasattr(val, 'layers') and hasattr(val.layers, '__len__')
) or (
    not key.startswith(u'_') and
    key not in ('In', 'Out') and
    not callable(val) and
    not inspect.isclass(val) and
    not inspect.ismodule(val)
)
]
'''


class FailedToGetWorkspaceVariable(Exception):
    
    def __repr__(self):
        
        return _('Failed to get workspace variable')


class TransparentJupyterWidget(RichJupyterWidget, BaseSubcomponent):

    def __init__(self, jupyter_console):

        self._name = jupyter_console.name
        self._jupyter_console = jupyter_console
        super(TransparentJupyterWidget, self).__init__(jupyter_console)
        self.setup(jupyter_console)
        self.executed.connect(self._on_executed)
        self.executing.connect(self._on_executing)
        self._last_executed = None
        # It is possible for _on_executing() to be called again, before
        # _on_executed() is called. We therefore keep track of how many times
        # _on_executing() has been called, and only release the busy status
        # when this counter is set back to 0.
        self._executing_counter = 0
        
    def execute(self, *args, **kwargs):
        
        self._last_executed = args[0] if args else kwargs.get('source', None)
        super(TransparentJupyterWidget, self).execute(*args, **kwargs)

    def _on_executing(self):

        self._executing_counter += 1
        self._jupyter_console.set_busy(True)
        self.extension_manager.fire('jupyter_execute_start')

    def _on_executed(self):
        """Is called when the kernel is done executing code. Oddly, this can
        happen just before the last bit of output is captured. Therefore, we
        don't react right away, but call the finish() function 100 ms later.
        """
        def finish():
            if self._executing_counter:
                self._executing_counter -= 1
            if self._executing_counter:
                return
            self._jupyter_console.set_busy(False)
            self.extension_manager.fire(
                u'workspace_update',
                name=self._name,
                workspace_func=self.get_workspace_globals
            )
            self.extension_manager.fire('jupyter_execute_finished')
        QTimer.singleShot(100, finish)
        
    def running(self):
        
        return self._executing_counter > 0
    
    def reset(self, clear=False):
        
        super(TransparentJupyterWidget, self).reset(clear=clear)
        self._executing_counter = 0
        self._jupyter_console.set_busy(False)
        
    def _handle_display_data(self, msg):
        """Reimplemented to handle communications between the figure explorer
        and the kernel. Inspired by Spyder's `figurebrowser.py`.
        """
        img = None
        data = msg['content']['data']
        if 'image/svg+xml' in data:
            fmt = 'image/svg+xml'
            img = data['image/svg+xml']
        elif 'image/png' in data:
            # PNG data is base64 encoded as it passes over the network
            # in a JSON structure so we decode it.
            fmt = 'image/png'
            img = decodebytes(data['image/png'].encode('ascii'))
        elif 'image/jpeg' in data and self._jpg_supported:
            fmt = 'image/jpeg'
            img = decodebytes(data['image/jpeg'].encode('ascii'))
        if img is not None:
            self.extension_manager.fire(
                'jupyter_execute_result_image',
                img=img,
                fmt=fmt,
                code=self._last_executed
            )
        return super(TransparentJupyterWidget, self)._handle_display_data(msg)
        
    def _append_plain_text(self, text, *args, **kwargs):
        
        super(TransparentJupyterWidget, self)._append_plain_text(
            text,
            *args,
            **kwargs
        )
        if not self._executing_counter:
            return
        self.extension_manager.fire('jupyter_execute_result_text', text=text)
        if 'Traceback (most recent call last)' in text:
            self.extension_manager.fire('jupyter_exception_occurred')


class OutprocessJupyterWidget(TransparentJupyterWidget):

    """Makes the Python workspace of a Jupyter console with a kernel running in
    a different process accessible.
    """

    def __init__(self, jupyter_console):

        self._user_expressions = {}
        super(OutprocessJupyterWidget, self).__init__(jupyter_console)

    def _handle_execute_reply(self, msg):

        self._user_expressions = msg.get(
            u'content',
            {}
        ).get(u'user_expressions', {})
        return super(OutprocessJupyterWidget, self)._handle_execute_reply(msg)
        
    def _silent_execute(self, expr):
        
        key = str(uuid.uuid4())
        self._kernel_client.execute(
            u'import os, inspect',
            silent=True,
            user_expressions={
                key: expr
            }
        )
        for _ in range(100):
            if key in self._user_expressions:
                break
            time.sleep(0.01)
            QApplication.processEvents()
        else:
            return {u'no reply': None}
        reply = self._user_expressions[key].get(u'data', {}).get(
            u'text/plain',
            u'{"invalid reply": None}'
        )
        try:
            return ast.literal_eval(reply)
        except (ValueError, SyntaxError):
            return {u'cannot eval reply': None}

    def get_workspace_globals(self):

        return self._silent_execute(GLOBAL_EXPR)

    def list_workspace_globals(self):

        return self._silent_execute(LIST_GLOBAL_EXPR)

    def set_workspace_globals(self, global_dict):

        code = ['from pickle import loads']
        for var, val in global_dict.items():
            if (
                var.startswith('_') or
                callable(val) or
                inspect.isclass(val) or
                inspect.ismodule(val)
            ):
                continue
            try:
                blob = pickle.dumps(val)
            except (pickle.PicklingError, TypeError):
                pass
            else:
                code.append('{} = loads({})'.format(var, repr(blob)))
        self._kernel_client.execute(';'.join(code), silent=True)

    def get_workspace_variable(self, name):

        key = str(uuid.uuid4())
        self._kernel_client.execute(
            u'import pickle',
            silent=True,
            user_expressions={
                key: u'pickle.dumps({})'.format(name)
            }
        )
        for _ in range(200):
            if key in self._user_expressions:
                break
            time.sleep(0.05)
            QApplication.processEvents()
        else:
            return FailedToGetWorkspaceVariable()
        reply = self._user_expressions[key].get(
            u'data',
            {}
        ).get(u'text/plain', None)
        if reply is None:
            return FailedToGetWorkspaceVariable()
        try:
            return pickle.loads(eval(reply))
        except (
            TypeError,
            pickle.UnpicklingError,
            RecursionError,
            ModuleNotFoundError
        ):
            return FailedToGetWorkspaceVariable()


class InprocessJupyterWidget(TransparentJupyterWidget):

    """Makes the Python workspace of a Jupyter console with a kernel running in
    the same process accessible.
    """

    def __init__(self, parent):

        self._jupyter_console = parent
        super(InprocessJupyterWidget, self).__init__(parent)

    def get_workspace_globals(self):

        return {
            key: (
                json.dumps(val, default=lambda x: u'<no preview>'),
                val.__class__.__name__,
                len(val) if hasattr(val, '__len__')
                else len(val.layers) if hasattr(val, 'layers') and hasattr(val.layers, '__len__')
                else '<na>'
            )
            for key, val
            in self._kernel_manager.kernel.shell.user_global_ns.copy().items()
            if (
                hasattr(val, 'layers') and hasattr(val.layers, '__len__')
            ) or (
                not key.startswith(u'_') and
                key not in ('In', 'Out') and
                not callable(val) and
                not inspect.isclass(val) and
                not inspect.ismodule(val)
            )
        }
        
    def list_workspace_globals(self):
        
        return [
            key
            for key, val
            in self._kernel_manager.kernel.shell.user_global_ns.copy().items()
            if (
                hasattr(val, 'layers') and hasattr(val.layers, '__len__')
            ) or (
                not key.startswith(u'_') and
                key not in ('In', 'Out') and
                not callable(val) and
                not inspect.isclass(val) and
                not inspect.ismodule(val)
            )
        ]

    def set_workspace_globals(self, global_dict):

        self._kernel_manager.kernel.shell.push(global_dict)

    def get_workspace_variable(self, name):

        return self._kernel_manager.kernel.shell.user_global_ns.get(name, None)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import sys
import logging
import logging.handlers


MAXBYTES = 10 * 1024 ** 2
BACKUPCOUNT = 1


class OSLogger:

    r"""Implements application logging through the Python logging module."""
    def start(self, name=u'default'):
        r"""When the module is imported, an uninitialized singleton instance of
        OSLogger is created. OSLogger.start() is then called to Initialize the
        actual logger.
        """
        self._name = name
        self._formatter = logging.Formatter(
            u'[%(asctime)s:%(module)s:%(lineno)s:%(levelname)s] %(message)s'
        )
        self._logger = logging.getLogger(name)
        self._logger.propagate = False
        self.add_handler(self.StreamHandler(sys.stdout))
        if self.debug_mode:
            self._logger.setLevel(logging.DEBUG)
            try:
                self.add_handler(self.RotatingFileHandler())
            except PermissionError:
                self.error(u'failed to set RotatingFileHandler')
        else:
            self._logger.setLevel(logging.INFO)

    def RotatingFileHandler(self, path=None, level=logging.DEBUG):
        r"""A factory that returns an instance of a RotatingFileHandler.

        Parameters
        ----------
        path, optional
            A logfile path or `None` to use the default path.
        level, optional
            A logging level.

        Returns
        -------
        A RotatingFileHandler.
        """
        if path is None:
            from libopensesame import misc
            path = os.path.join(
                misc.home_folder(),
                u'.opensesame',
                u'debug.log'
            )
        self.info('debug info in %s' % path)
        h = logging.handlers.RotatingFileHandler(
            path,
            maxBytes=MAXBYTES,
            backupCount=BACKUPCOUNT,
            encoding=u'utf-8'
        )
        h.setLevel(level)
        h.setFormatter(self._formatter)
        return h

    def StreamHandler(self, stream, level=logging.INFO):
        r"""A factory that returns an instance of a StreamHandler.

        Parameters
        ----------
        stream, optional
            A file-like object.
        level, optional
            A logging level.

        Returns
        -------
        A StreamHandler.
        """
        h = logging.StreamHandler(stream=stream)
        h.setLevel(level)
        h.setFormatter(self._formatter)
        return h

    def add_handler(self, handler):
        r"""Adds a handler."""
        self._logger.addHandler(handler)

    def remove_handler(self, handler):
        r"""Removes a handler."""
        self._logger.removeHandler(handler)

    @property
    def debug_mode(self):

        return '--debug' in sys.argv or '-d' in sys.argv

    @property
    def debug(self):

        return self._logger.debug

    @property
    def info(self):

        return self._logger.info

    @property
    def warning(self):

        return self._logger.warning

    @property
    def error(self):

        return self._logger.error

    @property
    def critical(self):

        return self._logger.critical

    @property
    def name(self):

        return self._logger.name

    @name.setter
    def name(self, name):

        self._logger.name = name

    @property
    def started(self):

        return hasattr(self, u'_logger')


# A singleton instance
oslogger = OSLogger()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.dialogs.base_dialog import BaseDialog


class TextInput(BaseDialog):

    r"""A simple text-input dialog."""
    def __init__(self, main_window, msg=None, content=u'', validator=None):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main window object.
        msg
            A text message.
        content, optional
            The starting content.
        validator, optional
            A function to validate the input. This function should take a
            single str as argument, and return a bool.
        """
        super().__init__(main_window, ui=u'dialogs.text_input_dialog')
        if msg is not None:
            self.ui.label_message.setText(msg)
        self._validator = validator
        self.ui.textedit_input.setPlainText(content)
        self.adjustSize()

    def get_input(self):
        r"""Gets text input.

        Returns
        -------
        A string with text input or None of the dialog was not accepted.
        """
        while True:
            if self.exec_() != QtWidgets.QDialog.Accepted:
                break
            s = self.ui.textedit_input.toPlainText()
            if self._validator is None or self._validator(s):
                return s
        return None


# Alias for backwards compatibility
text_input = TextInput



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import time
import functools
import traceback
from qtpy import QtWidgets, QtCore
from openexp import resources
from libopensesame.misc import snake_case, camel_case
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context


class BaseExtension(BaseSubcomponent):

    r"""A base class for GUI extensions."""
    extension_filter = None
    preferences_ui = None

    def __init__(self, main_window, info=None):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main-window object.
        info, optional
            deprecated
        """
        oslogger.debug(u'creating %s' % self.name())
        if info is not None:
            oslogger.warning('the info keyword is deprecated')
        if main_window.options.profile:
            # Use a special fire() function that keeps track of durations
            self._event_durations = {}
            self.fire = self._fire_and_time
        self._ = translation_context(snake_case(self.name()),
                                     category='extension')
        self.setup(main_window)
        if self.name() in self.unloaded_extension_manager:
            self._unloaded_extension = \
                self.unloaded_extension_manager[self.name()]
        else:
            self._unloaded_extension = \
                self.unloaded_extension_manager[snake_case(self.name())]
        self.register_ui_files()
        self.register_config()
        # Store of aliases so that we can still call the extension by its old
        # name if it has been renamed, e.g. OpenSesameIDE (old) for
        # opensesame_ide (new).
        name = self.name()
        self.aliases = [name]
        self.aliases += self.extension_attribute('aliases', [])
        if name.islower():
            self.aliases.append(camel_case(name))
        else:
            self.aliases.append(snake_case(name))
        try:
            qm_path = self.ext_resource('locale/' + main_window.locale + '.qm')
        except Exception as e:
            pass
        else:
            if main_window.translators:
                from qtpy.QtCore import QCoreApplication
                oslogger.info('installing translator {}'.format(qm_path))
                self._translator = main_window.translators.pop()
                self._translator.load(qm_path)
                QCoreApplication.installTranslator(self._translator)
            else:
                oslogger.warning('no translators objects available')
        self.create_action()

    @staticmethod
    def as_thread(wait):
        r"""A decorator to execute events in a separate thread, optionally
        after a delay.
        """
        def inner(fnc):
            def innermost(self, *args, **kwargs):
                QtCore.QTimer.singleShot(wait,
                    functools.partial(fnc, self, *args, **kwargs))
            return innermost
        return inner

    @property
    def console(self):
        return self.main_window.console

    @property
    def menubar(self):
        return self.main_window.menuBar()

    @property
    def toolbar(self):
        return self.main_window.ui.toolbar_main

    @property
    def tabwidget(self):
        return self.main_window.ui.tabwidget

    @property
    def statusbar(self):
        return self.main_window.statusBar()

    @property
    def set_busy(self):
        return self.main_window.set_busy
    
    @property
    def extension_attribute(self):
        return self._unloaded_extension.attribute
    
    @property
    def extension_folder(self):
        return self._unloaded_extension.folder
    
    def folder(self):
        oslogger.warning('BaseExtension.folder() is deprecated, use '
                         'BaseExtension.extension_folder instead')
        return self._unloaded_extension.folder

    def label(self):
        r"""Gives the label that is used for the menu and toolbar entry.
        Normally, this is specified in info.json, but you can override this
        function to implement custom logic.

        Returns
        -------
        unicode, NoneType
            A label text or None.
        """
        if not hasattr(self, '_label'):
            label = self.extension_attribute('label', None)
            self._label = None if label is None else self._(label)
        return self._label

    def tooltip(self):
        r"""Gives the tooltip that is used for the menu and toolbar entry.
        Normally, this is specified in info.json, but you can override this
        function to implement custom logic.

        Returns
        -------
        unicode, NoneType
            A tooltip text.
        """
        if not hasattr(self, '_tooltip'):
            tooltip = self._(self.extension_attribute(u'tooltip', None))
            if not tooltip:
                self._tooltip = None
            else:
                shortcut = self.shortcut()
                if not shortcut:
                    self._tooltip = tooltip
                else:
                    self._tooltip = '{} ({})'.format(tooltip, shortcut)
        return self._tooltip

    def checkable(self):
        r"""Indicates whether the extension action is checkable or not.
        Normally, this is specified in info.json, but you can override this
        function to implement custom logic.

        Returns
        -------
        bool
        """
        return self.extension_attribute(u'checkable', False)

    def set_checked(self, checked):
        r"""Sets the checked status of the action. If there is no action, or if
        the action is not checkable, this is silently ignored.

        Parameters
        ----------
        checked : bool
            The checked status.
        """
        if self.action is None or not self.checkable():
            return
        self.action.setChecked(checked)

    def set_enabled(self, enabled=True):
        r"""Enables/ disables the action. If there is no action, this is
        silently ignored.

        Parameters
        ----------
        enabled : bool
            The enabled status.
        """
        if self.action is None:
            return
        self.action.setEnabled(enabled)

    def icon(self):
        r"""Gives the name of the icon that is used for the menu and toolbar
        entry. Normally, this is specified in info.json, but you can override
        this function to implement custom logic.

        Returns
        -------
        unicode
            The name of an icon.
        """
        return self.extension_attribute(u'icon', u'applications-utilities')

    def shortcut(self):
        r"""Gives the keyboard shortcut that activates the extension. Normally,
        this is specified in info.json, but you can override this function to
        implement custom logic. A shortcut only works if the extension has
        either a toolbar or menu entry.

        Returns
        -------
        unicode
            The keyboard shortcut.
        """
        return self.extension_attribute(u'shortcut', None)

    def menu_pos(self):
        r"""Describes the position of the extension in the menu. Normally, this
        is specified in info.json, but you can override this function to
        implement custom logic.

        Returns
        -------
        tuple, NoneType
            A (submenu, menuindex, separator_before, separator_after) tuple, or
            None if the extension has no menu entry.
        """
        menu_pos = self.extension_attribute(u'menu', None)
        if menu_pos is None:
            return None
        return (menu_pos.get(u'submenu', None),
                menu_pos.get(u'index', -1),
                menu_pos.get(u'separator_before', False),
                menu_pos.get(u'separator_after', False))

    def toolbar_pos(self):
        r"""Describes the position of the extension in the toolbar. Normally,
        this is specified in info.json, but you can override this function to
        implement custom logic.

        Returns
        -------
        tuple, NoneType
            An (index, separator_before, separator_after) tuple, or None if the
            extension has no toolbar entry.
        """
        toolbar_pos = self.extension_attribute(u'toolbar', None)
        if toolbar_pos is None:
            return None
        return (toolbar_pos.get(u'index', -1),
                toolbar_pos.get(u'separator_before', False),
                toolbar_pos.get(u'separator_after', False))

    def activate(self):
        r"""Is called when the extension is activated through a keyboard
        shortcut, or a menu/ toolbar click. Override this function to implement
        your extension's behavior.
        """
        pass

    def _activate(self):
        r"""A wrapper around [activate] to catch Exceptions."""
        try:
            self.activate()
        except Exception as e:
            self.notify(
                f'Extension {self.name()} misbehaved on activate '
                f'(see debug window for stack trace)',
                category='warning')
            self.console.write(e)
            traceback.print_exc()
            oslogger.error(
                f'Extension {self.name()} misbehaved on activate: {e}')

    def add_action(self, widget, action, index, separator_before,
                   separator_after):
        r"""Adds an action to a widget that supports actions (i.e. a QMenu or a
        QToolBar).

        Parameters
        ----------
        widget
            The widget to add the action to.
        action
            The action to add.
        index
            The index of the action.
        separator_before
            Indicates whether a separator should be added before the action.
        separator_action
            Indicates whether a separator should be added before the action.
        """
        if index >= len(widget.actions()) or index < -len(widget.actions()) \
                or index == -1:
            before = None
            widget.addAction(action)
        else:
            if index < 0:
                index += 1
            before = widget.actions()[index]
        if separator_after:
            if before is None:
                before = widget.addSeparator()
            else:
                before = widget.insertSeparator(before)
        if before is None:
            widget.addAction(action)
        else:
            widget.insertAction(before, action)
        if separator_before:
            widget.insertSeparator(action)

    def get_submenu(self, menu_text):
        r"""Gets the submenu that matches the menu text. If no match is found a
        new submenu is created.

        Parameters
        ----------
        menu_text
            The menu text. This should match an object in
            main_window.ui.menu_[menu_text].

        Returns
        -------
        QMenu
        """
        menu_name = u'menu_%s' % menu_text.lower()
        if not hasattr(self.main_window.ui, menu_name):
            menu_name = u'menu_custom_{}'.format(len(self.menubar.actions()))
            menu = QtWidgets.QMenu(menu_text)
            self.menubar.addMenu(menu)
            setattr(self.main_window.ui, menu_name, menu)
        return getattr(self.main_window.ui, menu_name)

    def register_ui_files(self):
        r"""Registers all .ui files in the extension folder so that they can be
        retrieved as extensions.[extension name].[ui name].
        """
        for path in os.listdir(self.extension_folder):
            if path.endswith(u'.ui'):
                basename = os.path.splitext(path)[0]
                resource_path = os.path.join(self.extension_folder, path)
                name = self.name()
                resources[f'extensions.{name}.{basename}'] = resource_path
                if not name.islower():
                    resources[f'extensions.{snake_case(name)}.{basename}'] = \
                        resource_path

    def qaction(self, icon, label, target, checkable=False, tooltip=None,
                shortcut=None):
        """Creates a QAction for an extension."""
        action = QtWidgets.QAction(
            self.theme.qicon(icon), label, self.main_window)
        action.triggered.connect(target)
        action.setCheckable(checkable)
        if tooltip is not None:
            action.setToolTip(tooltip)
        if shortcut is not None:
            action.setShortcuts([shortcut])
        return action

    def create_action(self):
        r"""Creates a QAction for the extension, and adds it to the menubar
        and/ or the toolbar.
        """
        label = self.label()
        shortcut = self.shortcut()
        if label is None and shortcut is None:
            self.action = None
            return
        self.action = self.qaction(self.icon(), label,
                                   self._activate, checkable=self.checkable(),
                                   tooltip=self.tooltip(), shortcut=shortcut)
        # Insert the action into the menu
        if self.menu_pos() is not None:
            submenu_text, index, separator_before, separator_after = \
                self.menu_pos()
            submenu = self.get_submenu(submenu_text)
            self.add_action(submenu, self.action, index, separator_before,
                            separator_after)
        # Insert the action into the toolbar
        if self.toolbar_pos() is not None:
            index, separator_before, separator_after = self.toolbar_pos()
            self.add_action(self.toolbar, self.action, index,
                            separator_before, separator_after)

    def register_config(self):
        r"""Registers the extension settings in the config object."""
        for setting, default in self.extension_attribute(u'settings', {}).items():
            if isinstance(default, dict):
                default = default.get(self.main_window.mode, u'')
            cfg.register(setting, default=default)

    def _settings_widget_from_ui(self, ui):
        r"""Generates a standard preferences widget based on a ui file.

        Parameters
        ----------
        ui
            A ui file.
        """
        from libqtopensesame.widgets.base_preferences_widget import (
            BasePreferencesWidget
        )
        return BasePreferencesWidget(self.main_window, ui)

    def settings_widget(self):
        r"""Creates a settings QWidget for the extension, or returns None if no
        settings have been defined. Override this function to implement a more
        fancy non-default settings menu.

        Returns
        -------
        A settings QWidget or None.
        """
        if self.preferences_ui is not None:
            return self._settings_widget_from_ui(self.preferences_ui)
        r = 10000000  # Maximumum range for spinbox widgets
        settings = self.extension_attribute('settings', None)
        if settings is None:
            return
        group = QtWidgets.QGroupBox(
            self._(u'Extension: %s') % self.name(),
            self.main_window
        )
        group.__advanced__ = True
        layout = QtWidgets.QFormLayout(group)
        self.settings_controls = {}
        for setting, default in settings.items():
            value = cfg[setting]
            label = QtWidgets.QLabel(self._(setting))
            if isinstance(default, bool):
                control = QtWidgets.QCheckBox(self.main_window)
                control.setChecked(bool(value))
                control.clicked.connect(self.apply_settings_widget)
            elif isinstance(default, int):
                try:
                    value = int(value)
                except:
                    value = default
                control = QtWidgets.QSpinBox(self.main_window)
                control.setRange(-r, r)
                control.setValue(value)
                control.editingFinished.connect(self.apply_settings_widget)
            elif isinstance(default, float):
                try:
                    value = float(value)
                except:
                    value = default
                control = QtWidgets.QDoubleSpinBox(self.main_window)
                control.setRange(-r, r)
                control.setValue(value)
                control.editingFinished.connect(self.apply_settings_widget)
            else:
                control = QtWidgets.QLineEdit(safe_decode(value))
                control.editingFinished.connect(self.apply_settings_widget)
            self.settings_controls[setting] = control
            layout.addRow(label, control)
        group.setLayout(layout)
        return group

    def apply_settings_widget(self):
        r"""Applies the settings widget. This function is called automatically
        when a default settings widget is created by [settings_widget].
        """
        
        for setting, default in self.extension_attribute(u'settings', {}).items():
            control = self.settings_controls[setting]
            if isinstance(default, bool):
                cfg[setting] = control.isChecked()
            elif isinstance(default, int) or isinstance(default, float):
                cfg[setting] = control.value()
            else:
                cfg[setting] = str(control.text())

    def ext_resource(self, resource):
        r"""Finds an extension resource, i.e. a file in the extension folder,
        and returns the full path to the resource. An FileNotFoundError is 
        raised if the resource does not exist.

        Parameters
        ----------
        resource : str, unicode
            The name of a resource.

        Returns
        -------
        unicode
            The full path to the resource.
        """
        path = os.path.join(self.extension_folder, u'locale', self.locale,
                            resource)
        if os.path.exists(path):
            return path
        path = os.path.join(self.extension_folder, resource)
        if not os.path.exists(path):
            raise FileNotFoundError(f'Extension resource not found: {path}')
        return path



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import shlex
import re
import codecs
import os
import yaml
import warnings
from distutils.version import StrictVersion
from libopensesame import metadata
from libopensesame.exceptions import InvalidConditionalExpression, \
    InvalidOpenSesameScript
from libopensesame.item_stack import item_stack_singleton
from libopensesame.py3compat import *


SQUARE_BRACKET_TEXT_DEPRECATION_WARNING = 'The square-bracket notation is ' \
    'deprecated and will be removed in future versions of OpenSesame. ' \
    'Use f-string notation instead.'
PYTHON_CONDITIONAL_DEPRECATION_WARNING = 'Starting a conditiional statement ' \
    'with an equals sign is no longer necessary and will be removed in ' \
    'future versions of OpenSesame.'
SQUARE_BRACKET_CONDITIONAL_DEPRECATION_WARNING = 'The square-bracket ' \
    'notation is deprecated and will be removed in future versions of ' \
    'OpenSesame. Use Python expressions instead.'


class Syntax:

    """The `syntax` class implement text operations, including those that are
    necessary for interpreting OpenSesame script.
    
    Parameters
    ----------
    experiment
        The experiment object.
    """
    def __init__(self, experiment):

        self.experiment = experiment
        # A regular expression to match keywords, which are characterized by:
        # - Start with a letter or underscore
        # - Consists only of letters, numbers, or underscores, except for
        # - The last character, which is an equals sign
        self.re_cmd = re.compile(r'\A[_a-zA-Z]+[_a-zA-Z0-9]*=')
        # A regular expression to detect whether a string corresponds to an
        # f-string. This is done by detecting the presence of a single
        # unescaped {.
        self.re_fstring = re.compile(r'(?<!{){(?!{)')
        # A regular expression to match [variables] in text. The first part
        # indictates that we don't match items preceded by a slash
        # (\[variables]). This introduces a problem in that the slash itself can
        # also be escaped. To deal with this we eat up any even number of
        # slashes (\\\\)* so that only sequences preceded by an odd number of
        # slashes are preserved.
        self.re_txt = re.compile(
            r'(?<!\\)(\\\\)*(\[[_a-zA-Z]+[_a-zA-Z0-9]*\])')
        # A regular expression to match inline Python statements, like so:
        # [=10*10]
        # [\[test\]]
        self.re_txt_py = re.compile(r'(?<!\\)(\\\\)*(\[=.*?[^\\]\])')
        # Catch single equals signs
        self.re_single_eq = re.compile(r'(?<![=!<>])(=)(?!=)')
        # Catch 'never' and 'always'
        self.re_never = re.compile(r'\bnever\b', re.I)
        self.re_always = re.compile(r'\balways\b', re.I)
        # Strict no-variables is used to remove all characters except
        # alphanumeric ones
        self.re_sanitize_strict_novars = re.compile(r'[^\w]')
        # Strict with variables is used to remove all characters except
        # alphanumeric ones and [] signs that indicate variables
        self.re_sanitize_strict_vars = re.compile(r'[^\w\[\]]')
        # Matches valid variable names
        self.re_valid_var_name = re.compile(r'\A[_a-zA-Z]+[_a-zA-Z0-9]*\Z')
        # Loose is used to remove double-quotes, slashes, and newlines
        self.re_sanitize_loose = re.compile(r'[\n\"\\]')
        # Unsanitization is used to replace U+XXXX unicode notation
        self.re_from_ascii = re.compile(r'U\+([A-F0-9]{4})')
        self.re_front_matter = re.compile(r'---(?P<info>.*?)---', re.S)

    def auto_type(self, val, allow_bool=False, allow_none=False):
        """Casts a value to its best-fitting type, i.e. float, int, or
        unicode.

        Parameters
        ----------
        val
            A value of any type.
        allow_bool: bool
            Indicates whether 'True' and 'False' should be returned as bool
        allow_none: bool
            Indicates whether 'None' should be returned as None

        Returns
        -------
        An auto-typed value.
        """
        stripped_val = val
        if allow_bool and stripped_val == 'True':
            return True
        if allow_bool and stripped_val == 'False':
            return False
        if allow_none and stripped_val == 'None':
            return None
        try:
            f = float(val)
            # No flanking whitespace
            if isinstance(val, str):
                assert(val == stripped_val)
        except:
            return safe_decode(val, errors=u'ignore')
        if int(f) == f:
            return int(f)
        return f

    def parse_front_matter(self, s):
        r"""Parses the YAML front matter at the start of the script.

        Parameters
        ----------
        s : str
            The script.

        Returns
        -------
        dict
            A (d, s) tuple, where d is a dict with the front matter, and and s
            is the script with the front matter stripped off.
        """
        m = self.re_front_matter.match(s)
        try:
            d = safe_yaml_load(m.group(u'info'))
            s = s[len(m.group(0)):]
        except:
            d = {}
        # The API can be stored in a few ways:
        # - Not at all, which was the case in <= v2.9
        # - As a single value, in which the case the minor version becomes 0
        # - As a float, which is seen as a major.minor version
        if u'API' not in d:
            d['API'] = StrictVersion(u'1.0')
        elif isinstance(d['API'], int):
            d['API'] = StrictVersion(u'%d.0' % d['API'])
        elif isinstance(d['API'], float):
            d['API'] = StrictVersion(str(d['API']))
        return d, s

    def generate_front_matter(self):
        r"""Generates YAML front matter. Any existing front matter is copied,
        except for the `OpenSesame`, `API`, and `Platform` fields, which are
        updated.

        Returns
        -------
        str
            A front matter string.
        """
        self.experiment.front_matter.update({
            'OpenSesame': safe_str(metadata.__version__),
            'API': metadata.api,
            'Platform': os.name
        })
        front_matter = self.experiment.front_matter.copy()
        front_matter[u'API'] = metadata.api.version[0] \
            if metadata.api.version[1] == 0 \
            else float(str(metadata.api))
        return u'---\n%s---\n' % safe_decode(yaml.dump(
            front_matter, default_flow_style=False,
            allow_unicode=True))

    def split(self, s):
        r"""A unicode-safe bash-style split function. This is a wrapper around
        shlex.split() which is not unicode-safe on Python 2.

        Parameters
        ----------
        s : str, unicode
            The string to split.

        Returns
        -------
        list
            The string split into a list.
        """
        try:
            return shlex.split(s)
        except Exception as e:
            raise InvalidOpenSesameScript(f'Failed to parse line "{s}". Is '
                                          f'there a closing quotation '
                                          f'missing?')

    def parse_cmd(self, cmd):
        """Parses OpenSesame command strings, which consist of a command,
        optionally followed by arguments, optionally followed by keywords.

        For example:

            widget 0 0 1 1 label text="Demo text"

        Here, `widget` is the command, `0` through `label` are arguments,
        and `text` is a keyword.

        Parameters
        ----------
        cmd: str
            The command string to parse.

        Returns
        -------
        tuple
            A (command, arglist, kwdict) tuple.
        """
        l = self.split(cmd)
        if len(l) == 0:
            return None, [], {}
        cmd = l[0]
        arglist = []
        kwdict = {}
        for s in l[1:]:
            m = self.re_cmd.match(s)
            if m is not None:
                arg = s[:m.end()-1]
                val = s[m.end():]
                kwdict[arg] = self.auto_type(val)
            else:
                arglist.append(self.auto_type(s))
        return cmd, arglist, kwdict

    def create_cmd(self, cmd, arglist=[], kwdict={}):
        r"""Generates a command string.

        Parameters
        ----------
        cmd : str
            The command.
        arglist : list, optional
            A list of compulsory order-specified arguments.
        kwdict : dict, optional
            A dict of optional named keywords.
        """
        l = [cmd]
        for arg in arglist:
            l.append(self.safe_wrap(arg))
        for key, val in sorted(kwdict.items()):
            l.append(u'%s=%s' % (key, self.safe_wrap(val)))
        return u' '.join(l)
        
    def contains_variables(self, txt):
        """
        Returns
        -------
        bool
            True if txt contains any variables references, and False otherwise.
        """
        return any([self.re_fstring.search(txt), self.re_txt.search(txt),
                    self.re_txt_py.search(txt)])

    def eval_text(self, txt, round_float=False, var=None, include_local=False):
        """Evaluates variables and inline Python in a text string.

        Examples
        --------
        The resolution is [width] by [height] pixels
        This evaluates to 100: [=10x10]

        Parameters
        ----------
        txt:
             The string to evaluate. If the input is not a string, then the
             value will be returned unmodified.
        round_float: bool, optional
            Indicates whether floating point values should be rounded or not.
        var: VarStore or None, optional
            A deprecated option that specific var store for evaluating the
            string. This does not apply to f-strings.
        include_local : bool, optional
            If True, the variable store of the current item is merged into the
            Python workspace. This allows items to evaluate f-strings that
            take into account the item's local variables. For old-style
            [variables], this means that the local variable store is evaluated.

        Returns
        -------
        The evaluated string, or the input value for non-string input.
        """

        if not isinstance(txt, str):
            return txt
        if self.re_fstring.search(txt):
            return self.experiment.python_workspace.eval_fstring(
                txt, include_local=include_local)
        
        def get_escape_sequence(m):
            return u'' if m.group(1) is None \
                else m.group(1)[:len(m.group(1))//2]
        if var is None:
            # If we want to include local variables, we need to get the current
            # item from the stack
            if include_local:
                item_name, phase = item_stack_singleton[-1]
                var = self.experiment.items[item_name].var
            else:
                var = self.experiment.var
        if round_float:
            float_template = u'%%.%sf' % var.round_decimals
        while True:
            m = self.re_txt.search(txt)
            if m is None:
                break
            warnings.warn(SQUARE_BRACKET_TEXT_DEPRECATION_WARNING,
                          category=DeprecationWarning)
            varname = m.group(2)[1:-1]
            val = var.get(varname)
            if round_float and isinstance(val, float):
                val = float_template % val
            else:
                val = safe_decode(val)
            txt = txt[:m.start(0)] + get_escape_sequence(m) + val \
                + txt[m.end(0):]
        # Detect Python inlines [=10*10]
        while True:
            m = self.re_txt_py.search(txt)
            if m is None:
                break
            warnings.warn(SQUARE_BRACKET_TEXT_DEPRECATION_WARNING,
                          category=DeprecationWarning)
            py = self.unescape(m.group(2)[2:-1])
            val = self.experiment.python_workspace._eval(py)
            txt = txt[:m.start(0)] + get_escape_sequence(m) + safe_decode(val) \
                + txt[m.end(0):]
        return self.unescape(txt)

    def quotable_symbol(self, s):
        """
        Returns
        -------
        bool
            True if s is a symbol that should be quoted in a conditional
            statement, and False otherwise.
        """
        # Don't quote operators
        if s in [u'not', u'or', 'and']:
            return False
        # Don't quote special keywords
        if s.lower() in [u'always', u'never']:
            return False
        # Don't quote numeric values
        try:
            int(s)
            return False
        except:
            return True

    def compile_cond(self, cnd, bytecode=True):
        """Compiles OpenSesame conditional statements.

        Examples
        --------
        [width] > 100
        =var.width > 100

        Parameters
        ----------
        cnd: str
            The conditional statement to compile.
        bytecode: bool, optional
            Indicates whether the conditional statement should be returned as
            bytecode (True) or a Python string (False).

        Returns
        -------
        str or bytecoode
            The conditional statement as a Python string or bytecode.
        """
        cnd = safe_decode(cnd)
        # Python conditions `=True` don't have to be evaluated
        if cnd.startswith('='):
            warnings.warn(PYTHON_CONDITIONAL_DEPRECATION_WARNING,
                          category=DeprecationWarning)
            cnd = cnd[1:]
        # The logic for determining whether a conditional expression
        # corresponds to an old-style expression is a bit convoluted, but the
        # idea is that old-style expressions either contain square brackets or
        # the literal 'always' or 'never' but only if there are not hashtags in
        # there, because in that case the always and never are likely suffixes
        # in a comment ('True  # always').
        elif self.re_txt.search(cnd) or (
                ('always' in cnd.lower() or 'never' in cnd.lower())
                 and '#' not in cnd
            ):
            warnings.warn(SQUARE_BRACKET_CONDITIONAL_DEPRECATION_WARNING,
                          category=DeprecationWarning)
            # Quote all non-quoted symbols. This can be probably be done through
            # a clever regular expression, but for now we use a simple loop.
            while True:
                in_quote = None
                in_var = False
                symbol_start = None
                for i, ch in enumerate(cnd):
                    # Don't scan within quoted strings
                    if in_quote is None:
                        if ch in u'"\'':
                            in_quote = ch
                            continue
                    elif ch == in_quote:
                        in_quote = None
                        continue
                    if in_quote:
                        continue
                    # Don't scan within variable definitions
                    if ch == u'[':
                        in_var = True
                        continue
                    elif ch == u']':
                        in_var = False
                        continue
                    if in_var:
                        continue
                    # Detect symbols starts, i.e. the first alphanumeric
                    # character or underscore
                    if ch.isalnum() or ch == u'_':
                        if symbol_start is None:
                            symbol_start = i
                    # Detect symbol ends, i.e. the first non-alphanumeric
                    # character after an alphanumeric character or an
                    # underscore.
                    elif symbol_start is not None:
                        symbol = cnd[symbol_start:i]
                        symbol_range = symbol_start, i
                        symbol_start = None
                        if not self.quotable_symbol(symbol):
                            continue
                        cnd = cnd[:symbol_range[0]] + u'"' + symbol + u'"' \
                            + cnd[symbol_range[1]:]
                        break
                else:
                    # The for-else clause happens when the for loop was not
                    # broken. If no break occurred, then nothing was quoted,
                    # and we don't need to restart the for loop, i.e. we can
                    # break the infinite while. We only need to check that there
                    # was no symbol still being processed, in which case it
                    # needs to be quoted.
                    if symbol_start is not None:
                        symbol = cnd[symbol_start:]
                        if self.quotable_symbol(symbol):
                            cnd = cnd[:symbol_start] + u'"' + symbol + u'"'
                    break
            # Replace [variables] by var.variables
            while True:
                m = self.re_txt.search(cnd)
                if m is None:
                    break
                cnd = cnd[:m.start()] + u'var.%s' % m.group()[1:-1] \
                    + cnd[m.end():]
            # Replace single equals signs (=) by doubles (==)
            cnd = self.re_single_eq.sub(u'==', cnd)
            # Replace always and never words by True or False
            cnd = self.re_never.sub(u'False', cnd)
            cnd = self.re_always.sub(u'True', cnd)
        if bytecode:
            try:
                return compile(cnd, '<conditional statement>', 'eval')  # __ignore_traceback__
            except SyntaxError as e:
                raise InvalidConditionalExpression(
                    'Syntax error in conditional expression')
        return self.unescape(cnd)

    def unescape(self, s):
        r"""Unescapes square brackets in text.

        Parameters
        ----------
        s : str
            The text to unescape.

        Returns
        -------
        str
            The unescaped text.
        """
        s = safe_decode(s)
        return s.replace(u'\[', u'[').replace(u'\]', u']')

    def safe_wrap(self, s):
        """Wraps and escapes a text so that it can safely be embedded in a
        command string. For example:

        He said: "hi"

        becomes:

        "He said: \"hi\""

        Parameters
        ----------
        s: str
            The text to wrap.

        Returns
        -------
        str
            The wrapped text.
        """
        s = safe_decode(s).replace(u'\\', u'\\\\')
        try:
            float(s)
            assert(s == s.strip())  # No flanking whitespace
        except:
            t = s.replace(u'_', u'')
            if not t.isalnum() or not t.isascii():
                return u'"%s"' % s.replace(u'"', u'\\"')
        return s

    def sanitize(self, s, strict=False, allow_vars=True):
        r"""Removes invalid characters (notably quotes) from the string.

        Parameters
        ----------
        s    The string to be sanitized. This can be any type, but if it is not
            unicode, it will be coerced to unicode.
        strict : bool, optional
            If True, all except underscores and alphanumeric characters are
            stripped.
        allow_vars : bool, optional
            If True, square brackets are not sanitized, so you can use
            variables.

        Returns
        -------
        unicode
            A sanitized string.

        Examples
        --------
        >>> # Prints 'Universit Aix-Marseille'
        >>> print(self.sanitize('\"Université Aix-Marseille\"'))
        >>> # Prints 'UniversitAixMarseille'
        >>> print(self.sanitize('\"Université Aix-Marseille\""', strict=True))
        """
        s = safe_decode(s)
        if strict:
            if allow_vars:
                return self.re_sanitize_strict_vars.sub(u'', s)
            return self.re_sanitize_strict_novars.sub(u'', s)
        return self.re_sanitize_loose.sub(u'', s)

    def to_ascii(self, s, strict=False):
        r"""Converts all non-ASCII characters to U+XXXX notation, so that the
        resulting string can be treated as plain ASCII text.

        Parameters
        ----------
        s : unicode
            A unicode string to be sanitized
        strict : bool, optional
            If True, special characters are ignored rather than recoded.

        Returns
        -------
        str
            A regular Python string with all special characters replaced by
            U+XXXX notation or ignored (if strict).
        """
        if strict:
            _s = safe_encode(s, enc=u'ascii', errors=u'ignore')
        else:
            _s = codecs.encode(s, u'ascii', u'osreplace')
        _s = safe_decode(_s)
        return _s.replace(os.linesep, u'\n')

    def from_ascii(self, s):
        r"""Converts an ascii str with U+XXXX notation to actual Unicode.

        Parameters
        ----------
        s : str
            A plain-ascii string.

        Returns
        -------
        unicode
            A string with all U+XXXX characters converted to the corresponding
            unicode characters.
        """
        s = safe_decode(s)
        while True:
            m = self.re_from_ascii.search(s)
            if m is None:
                break
            s = s.replace(m.group(0), chr(int(m.group(1), 16)), 1)
        return s

    def valid_var_name(self, s):
        r"""Checks whether a string is a valid variable name.

        Parameters
        ----------
        s : str
            The string to check.

        Returns
        -------
        True if s is a variable valid name, False if not.
        """
        return re.match(self.re_valid_var_name, s)


def osreplace(exc):
    r"""A replacement function to allow opensame-style replacement of unicode
    characters.

    Parameters
    ----------
    exc : UnicodeEncodeError

    Returns
    -------
    tuple
        A (replacement, end) tuple.
    """
    _s = u''
    for ch in exc.object[exc.start:exc.end]:
        _s += u'U+%.4X' % ord(ch)
    return _s, exc.end


codecs.register_error(u'osreplace', osreplace)
# Alias for backwards compatibility
syntax = Syntax



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue
from openexp._canvas._element.group import Group
from openexp.canvas_elements import Line, Ellipse


class FixDot(Group):

    def __init__(self, canvas, x=None, y=None, style=u'default', **properties):

        if x is None:
            x = 0
        if y is None:
            y = 0
        h = 2
        if u'large' in style:
            s = 16
        elif u'medium' in style or style == u'default':
            s = 8
        elif u'small' in style:
            s = 4
        else:
            raise InvalidValue(u'Unknown style: %s' % self.style)
        if u'open' in style or style == u'default':
            elements = [
                Ellipse(x-s, y-s, 2*s, 2*s, fill=True, **properties)
                .construct(canvas),
                Ellipse(x-h, y-h, 2*h, 2*h, fill=True,
                        color=canvas.background_color.colorspec,
                        **{key: val for key, val in properties.items()
                           if key != u'color'}).construct(canvas)
            ]
        elif u'filled' in style:
            elements = [Ellipse(
                x-s, y-s, 2*s, 2*s, fill=True, **properties)
                .construct(canvas)]
        elif u'cross' in style:
            elements = [
                Line(x, y-s, x, y+s, **properties).construct(canvas),
                Line(x-s, y, x+s, y, **properties).construct(canvas)
            ]
        else:
            raise InvalidValue(u'Unknown style: %s' % self.style)
        properties = properties.copy()
        properties.update({'x': x, 'y': y, u'style': style})
        Group.__init__(self, canvas, elements, **properties)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from libopensesame.experiment import Experiment as ExperimentRuntime
from libopensesame.base_python_workspace import BasePythonWorkspace
import opensesame_plugins
from libqtopensesame.misc.qtplugin_manager import QtPluginManager
from libqtopensesame.misc.qtitem_store import QtItemStore
from libqtopensesame.misc.qtsyntax import QtSyntax
from qtpy import QtCore, QtWidgets, QtGui
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'experiment', category=u'item')


class Experiment(ExperimentRuntime):

    """Contains various GUI controls for the experiment"""
    def __init__(self, main_window, name='experiment', string=None,
                 pool_folder=None, experiment_path=None, resources={}):
        r"""Constructor. The experiment is created automatically be OpenSesame
        and you will generally not need to create it yourself.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        name : str, unicode, optional
            The name of the experiment.
        string : str, unicode, NoneType, optional
            A string containing the experiment definition, the name of an
            OpenSesame experiment file, or `None` to create a blank experiment.
        pool_folder : str, unicode, NoneType, optional
            A specific folder to be used for the file pool, or `None` to use a
            new temporary folder.
        experiment_path : str, unicode, NoneType, optional
            The path of the experiment file. This will need to be specified
            even if a filename was passed using the `string` keyword.
        resources : dict, optional
            A dictionary with names as keys and paths as values. This serves as
            a look-up table for resources.
        """
        self.main_window = main_window
        self.ui = self.main_window.ui
        self.unused_items = []
        self.core_items = [
            u"loop", u"sequence", u"sketchpad", u"feedback", u"sampler",
            u"synth", u"keyboard_response", u"mouse_response", u"logger",
            u"inline_script"
        ]
        self.items = QtItemStore(self)
        self._syntax = QtSyntax(self)
        self._plugin_manager = QtPluginManager(opensesame_plugins)
        super().__init__(name, string, pool_folder,
                         experiment_path=experiment_path, resources=resources,
                         fullscreen=None,
                         workspace=BasePythonWorkspace(self))

    @property
    def overview_area(self):
        return self.main_window.ui.itemtree

    @property
    def default_title(self):
        return _(u'New experiment')

    @property
    def default_description(self):
        return _(u'Default description')

    def module_container(self):
        """
        Specifies the module that is used to get items from.

        Returns:
        u'libqtopensesame.items'
        """
        return u'libqtopensesame.items'
        
    def item_prefix(self):
        """
        A prefix for the plug-in classes, so that [prefix][plugin] class is used
        instead of the [plugin] class.
        """
        return u'qt'

    def build_item_tree(self, toplevel=None, items=[], max_depth=-1,
                        select=None):
        r"""Builds the overview area for the full experiment.

        Parameters
        ----------
        toplevel, optional
            The toplevel widget.
        items, optional
            A list of items that have already been added, to prevent recursion.
        max_depth, optional
            The maximum depth of the tree.
        select, optional
            The selected item.
        """
        if self.overview_area.locked:
            return
        from libqtopensesame.widgets.tree_unused_items_item import \
            tree_unused_items_item
        from libqtopensesame.widgets.tree_general_item import tree_general_item
        self.treeitem_general = tree_general_item(self)
        self.treeitem_unused = tree_unused_items_item(self.main_window)
        fold_state = self.overview_area.get_fold_state()
        self.overview_area.clear()
        self.overview_area.insertTopLevelItem(0, self.treeitem_general)
        self.overview_area.insertTopLevelItem(1, self.treeitem_unused)
        self.overview_area.set_fold_state(fold_state)
        self.overview_area.scrollToTop()
        if select is not None:
            self.ui.itemtree.select_item(select)

    def rename(self, from_name, to_name):
        """
        Renames an item.
    
        Parameters
        ----------
        from_name : str
            The old name.
        to_name : str
            The new name.
        """
        if self.var.start == from_name:
            self.var.start = to_name

    def check_name(self, name):
        """
        Checks whether a given name is valid. Reasons for not being valid
        are invalid characters or a conflict with an existing name..

        Arguments:
        name	--	The name to check.

        Returns:
        True if the name is allowed, False otherwise.
        """
        if name.strip() == u'':
            return u'Empty names are not allowed.'
        if name.lower() in [item.lower() for item in self.items.keys()]:
            return u'An item with that name already exists.'
        if name != self.syntax.sanitize(name, strict=True, allow_vars=False):
            return u'Name contains special characters. Only alphanumeric characters and underscores are allowed.'
        return True

    def delete(self, item_name, item_parent=None, index=None):
        """Deletes an item.

        Parameters
        ----------
        item_name: str
            The name of the item to be deleted.
        item_parent: str or None, optional
            The parent item. (default=None)
        index: int or None, optional
            The index of the item in the parent sequence, if
        """
        if self.var.start == item_name:
            self.notify(
                u'You cannot delete the entry point of the experiment!')
            return
        for item in self.items:
            self.items[item].delete(item_name, item_parent, index)
        self.main_window.close_item_tab(item_name)


# Alias for backwards compatibility
experiment = Experiment



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import functools
from jupyter_client import kernelspec
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import BaseWidget
from libopensesame.oslogging import oslogger
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QMenu
from libqtopensesame.misc.translate import translation_context
from .constants import KERNEL_NAMES
_ = translation_context(u'JupyterConsole', category=u'extension')


class ConsoleCornerWidget(BaseWidget):

    def __init__(self, console_tabwidget, kwargs):

        super(ConsoleCornerWidget, self).__init__(console_tabwidget)
        self._console_tabwidget = console_tabwidget
        self._add_button = QPushButton()
        self._add_button.setIcon(self.main_window.theme.qicon(u'list-add'))
        self._add_button.setFlat(True)
        self._add_button.setMenu(self._kernel_menu())
        self._add_button.setToolTip(_(u'Start new console'))
        self._restart_button = QPushButton()
        self._restart_button.setIcon(
            self.main_window.theme.qicon(u'view-refresh')
        )
        self._restart_button.setFlat(True)
        self._restart_button.clicked.connect(self._restart)
        self._restart_button.setToolTip(_(u'Restart kernel'))
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._restart_button)
        self._layout.addWidget(self._add_button)
        self.setLayout(self._layout)

    def _add(self, kernel):

        self._console_tabwidget.add(kernel=kernel)

    def _installed_kernels(self):

        if cfg.jupyter_kernels:
            return safe_decode(cfg.jupyter_kernels).split(u';')
        kernelspec_manager = kernelspec.KernelSpecManager()
        kernels = list(kernelspec_manager.find_kernel_specs().keys())
        # Python 2 and 3 kernels apparently just start the same kernel that
        # corresponds to the active Python kernel. So if they are both
        # installed, we expose only one.
        if 'python2' in kernels and 'python3' in kernels:
            kernels.remove('python2')
        oslogger.debug('installed kernels: {}'.format(str(kernels)))
        return kernels

    def _kernel_menu(self):

        menu = QMenu(self.main_window)
        for kernel in self._installed_kernels():
            action = menu.addAction(KERNEL_NAMES.get(kernel, kernel))
            action.triggered.connect(
                functools.partial(self._add, kernel=kernel)
            )
        return menu

    def _restart(self):

        self._console_tabwidget.current.restart()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.item import Item
from libopensesame.exceptions import OSException


class QuestNotInitialized(OSException):
    """This exception is raised when a Quest staircase hasn't been initialized.
    """
    pass


class QuestStaircaseNext(Item):

    def reset(self):
        self.var.response_var = 'correct'
        self.var.quest_name = 'default'

    def run(self):
        resp = self.var.get(self.var.response_var)
        try:
            resp = int(resp)
        except (ValueError, TypeError):
            # Don't process non-float responses
            return
        if not hasattr(self.experiment, 'quest') or \
                self.var.quest_name not in self.experiment.quest:
            raise QuestNotInitialized(
                f'quest staircase {self.var.quest_name} has not been initialized with a quest_staircase_init')
        self.experiment.quest[self.var.quest_name].update(
            self.var.quest_test_value, resp)
        self.experiment.quest_set_next_test_value[self.var.quest_name]()
